-- pred prey ecology
--
-- with bounded growth. see "Differential Equations, Dynamical Systems, and
--	Linear Algebra", Hirsch and Smale.
--
-- single location for C partition, extremal right, nonintersecting
-- see "Linear Phase-Portrait Approximations for Nonlinear Hybrid Systems"
-- T. Henzinger and H. Wong-Toi

-- non intersecting lines, i.e. D/C > A/lam
-- Use M: Cx = D + mu y
--     L: A = By + lam x
-- rate is y' >= -C x' / B
-- y' >=0
-- init pt is (10,3)


define(D,26)
define(C,4)
define(B,3)
define(mu,2)
define(A, 20)
define(lam, 5)

--  rate of x in right region lies in
--	[(A- B ymax - lam xmax) xmax, (A - lam D/C) D/ C]
-- given ymax and xmax, assume to be 12, 7

define(MIN_right_dx, -732)
define(MAX_right_dx, -325/4)

var 
	x,      	-- prey
	y      		-- predator
	: analog;
	ymax		-- max value of y
	: parameter
	;
 -- -------------------------------------------------------------- *)

automaton p1
synclabs : ;

initially loc_right & x=10 & y=3;

loc loc_right: while C x>=D+mu y 
		wait {dy<= -C/B dx, dy>=0
--		        ,dx >= MIN_right_dx, dx <= MAX_right_dx 
			}

	when C x = D + mu y goto loc_middle;


loc loc_middle: while C x <= D + mu y  & A - B y - lam x<= 0 & y>=0 & x>=0 
		wait {dx <=0, dy<=0}
        when C x = D + mu y goto loc_right;
--	when y=0 goto loc_lower_middle;
	when A - B y - lam x = 0 goto loc_left;

loc loc_left: while A - B y - lam x >=0 & y>=0 & x >= 0
		wait {dx>=0, dy<=0}
	when A - B y - lam x = 0 goto loc_middle;


end

-- analysis commands


var
	init_reg, 
	reached, 
	final_y_reg, final_x_reg : region;

init_reg := loc[p1]=loc_right &  x=10 & y=3;

final_y_reg :=  y>=ymax; 



reached :=  reach forward from init_reg endreach;

prints "Unreachable values for maximum predator (var y) population";
print omit all locations
	hide non_parameters in reached
	   & final_y_reg
        endhide;



