-- temperature controller for reactor
--

-- hytech input file

-- need to test whether violating region is contained in the region for which
-- the bounds are met exactly, i.e to cope wi|h lack of < in automaton
-- which forces us to transition to shutdown when x1/x2 = T_reuse, not
-- only when < T_reuse

define(T_reuse,80)      -- min delay between reuse of rod
define(vr,34)           -- rate of increase with no rods present
define(v1,25)           -- rate of decrease in temp with rod 1 present
define(v2,10)           -- rate of decrease in temp with rod 2 present
define(tmax,1100)       -- maximal tolerable temperature
define(tmin,250)        -- minimal tolerable temperature

var
  x1,                   -- timer for rod 1
  x2 : stopwatch;       -- timer for rod 2
  t: analog;            -- temperature of coolant


automaton temp_controller
synclabs:;
initially no_rods & x1 = T_reuse & x2 = T_reuse;

loc no_rods: while t <= tmax wait {dt = vr, dx1=1,dx2=1}
	when t=tmax & x1>=T_reuse goto rod_1;
	when t=tmax & x2>=T_reuse goto rod_2;
	when t=tmax & x1<=T_reuse & x2<=T_reuse goto shutdown;

loc rod_1: while t >= tmin wait {dt = -v1, dx1=1, dx2=1}
	when t = tmin do {x1' = 0} goto no_rods;
	
loc rod_2: while t >= tmin wait {dt = -v2, dx1=1, dx2=1}
	when t = tmin do {x2' = 0} goto no_rods;
	
loc shutdown: while t=tmax wait {dt=1,dx1=1,dx2=1}

end

var reached, reached_viol, equality: region;

reached := reach forward from loc[temp_controller] = no_rods 
				& x1 = T_reuse
				& x2 = T_reuse
	  endreach;

reached_viol := reached & loc[temp_controller] = shutdown;

equality:= x1=T_reuse | x2=T_reuse;

if reached_viol <= equality
   then prints "System verified: temperature maintained within bounds";
   else prints "System not verified: temperature not maintained within bounds";
	print reached_viol;
endif;
