#include <malloc.h> 
#include "poly.H"

#define EXCH(mat, l1, l2, aux) \
  aux=mat.p[l1]; \
  mat.p[l1]=mat.p[l2]; \
  mat.p[l2]=aux; 


// g++ says this is already defined // #define NULL 0
#define TOP 2147483647

// #define MAXRAYS  2048
#define MAXRAYS 4096




 int mult_call_nr = 0;

int vec_norm_call_nr = 0;

// a global program control mode. set by flags.
extern int overflow_avoid_mode;


inline int ABS(int d) {
  if (d > 0) return d;
  else if (d == -2147483648) return 2147483647;
  else return -d; }

inline int sign_of(int m) {
  if (m<0) return -1;
  else if (m==0) return 0;
  else return 1;
}

inline int 
mymult(int m1, int m2) {
  int res = m1 * m2;
  int corr_res_sign;
  int sign_m1 = sign_of(m1);

  if (sign_m1 == 0) {
    corr_res_sign = 0;
  }
  else {
    int sign_m2 = sign_of(m2);
    if (sign_m2 == 0) {
      corr_res_sign = 0;
    }
    else {
      if (sign_m1 == sign_m2 ) {
	corr_res_sign = 1;
      }
      else {
	corr_res_sign = -1;
      }
    }
  }
      
  if (corr_res_sign != sign_of(res)) {
    cerr << "Library overflow error in multiplication\n";
//    cerr << "Mult call nr = " << mult_call_nr << "\n";
//    cerr << "Computes " << m1 << " * " << m2 << " = " << res << "\n";
    if (overflow_avoid_mode < 2) {
      cerr << "Try overflow avoidance modes, -o1, -o2\n";
    }
    exit (1);
    return(1); // to keep CC compiler happy
  }
  else {
    return res;
  }
}



inline void 
myinc(int *sum, int inc) {
  if (sign_of(*sum) != sign_of(inc)) {
    *sum += inc;
  }
  else {
    if (inc==0) {
      // both 0
      return;
    }
    else {
      // both pos or both neg
      int res = *sum+inc;
      if (sign_of(res) != sign_of(inc)) {
	cerr << "Library overflow error in addition\n";
	cerr << "Computes " << *sum << " + " << inc << " = " << res << "\n";
	exit (1);
      }
      else {
	*sum = res;
      }
    }
  }
}
	

void
out_of_mem_msg() {
    cerr << "Not enough memory\n" << flush;
    exit(1);
  }


/* ************************************************** */


vector::vector()
{
  size=0;
  p= (int*) 0;
};

vector::vector(int s)
{
  size=s;
  p= new int[size];
};

void vector::free()
{
  delete(p);
  size = 0;
};

void vector::print()
{
  (void) printf("%d \n", size);
  for (int i=0;i<size;i++) 
      (void) printf("%4d ", p[i]);
  printf("\n"); fflush(stdout);
};

// adding nonstrict info to constructors, HWT 9/26/95

matrix::matrix()
{
  nbrows = 0;
  nbcolumns = 0;
  p_init = 0;
  p = 0;
  nonstrict = 0;
};

matrix::matrix(int nr, int nc)
{
  nbrows=nr;
  nbcolumns=nc;
  p_init = new int[nr*nc];
  p = new (int*[nr]);
  int* q = p_init;
  for (int i=0;i<nbrows;i++) {
    p[i]=q;
    q=q+nbcolumns;
  }
  nonstrict = 0;
};

matrix::matrix(int nr)
{
  nbrows=nr;
  nbcolumns=0;
  p = new (int*[nr]);
  nonstrict = 0;
};

matrix::matrix(const matrix& m)
// copy constructor
{
   nbrows = m.nbrows;
   nbcolumns = m.nbcolumns;
   p = m.p;
   p_init = m.p_init;
   nonstrict = m.nonstrict;
};

matrix::matrix(int nr, int nc , int *tab)
{
  nbrows=nr;
  nbcolumns=nc;
  p_init = tab;
   p = new (int*[nr]);
  int* q = p_init;
  for (int i=0;i<nbrows;i++) {
    p[i]=q;
    q=q+nbcolumns;
  }
  nonstrict = 0;
}; /* Alloc_Matrix */

void matrix::free()
{
  nbrows = 0;
  nbcolumns = 0;
  delete(p_init);
  delete(p);
};

void matrix::operator=(const matrix& m)
{
   nbrows = m.nbrows;
   nbcolumns = m.nbcolumns;
   p = m.p;
   p_init = m.p_init;
   nonstrict = m.nonstrict;
};


vector* Read_Vector()
{
  vector* vec;
  int s;
  (void) scanf("%d", &s);
  vec = new vector(s);
  for(int i=0;i<s;i++) (void) scanf("%d", &(vec->p[i]));
  return vec;
}; /* Read_Vector */

matrix* Read_Matrix()
{
  matrix* mat;
  int nr, nc;

  (void) scanf("%d %d", &nr, &nc);
  mat = new matrix(nr,nc);
  for (int i=0;i<nr;i++)
    for (int j=0;j<nc;j++)
      (void) scanf("%d", &(mat->p[i][j]));
  return mat;
}; /* Read_Matrix */

void matrix::print(ostream& outf)
{

  (void) printf("{");
  for (int i=0;i<nbrows-1;i++) {
    (void) printf ("{ ");    
    for (int j=0;j<nbcolumns-1;j++)
        (void) printf("%d, ", p[i][j]);
    (void) printf("%d ", p[i][nbcolumns-1]);
    (void) printf("},");
  }
  (void) printf("{ ");    
  for (int j=0;j<nbcolumns-1;j++)
      (void) printf("%d, ", p[nbrows-1][j]);
  (void) printf("%d ", p[nbrows-1][nbcolumns-1]);
  (void) printf("}");
  (void) printf("}\n");

   if (nonstrict) {
     printf("Nonstrict\n");
   }
  else {
    printf("Strict: \n");
  }
  fflush(stdout);
}; /* Write_Matrix */

void matrix::print(int rows, ostream& outf)
{
  (void) printf("%d %d\n", nbrows, nbcolumns);
  for (int i=0;i<rows;i++) {
    for (int j=0;j<nbcolumns;j++)
      (void) printf("%4d ", p[i][j]);
    (void) printf("\n");
  }
  fflush(stdout);

}; /* Write_Matrix */


void Gcd(int a, int b, int *g)
{
  int aux;
  if ((a==b) && (b==0)) *g=0;
  else {
    a=ABS(a);
    b=ABS(b);
    while (a) { aux=b%a; b=a; a=aux; }
    *g=b;
  }
}; /* Gcd */


/* finds the minimum (absolute value) of vector vec */
void Vector_Min_Not_0(vector& vec, int& min, int& index)
{
  int m=0, size, i, ind, aux;

  size=vec.size;
  ind= -1;
  for (i=0;i<size;i++)
    if (vec.p[i]!=0)
      if ( (m>(aux=ABS(vec.p[i]))) || (m==0) ) { ind=i; m=aux; }
  if (ind== -1) min=1;
  else min=m;
  index=ind;
}; /* Vector_Min_Not_0 */


/* computes the gcd of all components of vec */
void Vector_Gcd(int* vec, int size, int& g)
{
  vector vec_A(size);
  int min, index, *p, not_all_zero, i;
  
  p=vec;
  for (i=0;i<size;i++)
    vec_A.p[i]= *p++;
  do {
    Vector_Min_Not_0(vec_A, min, index);
    if (min==1) break;
    not_all_zero=false;
    for (i=0;i<size;i++)
      if (i!=index) {
        vec_A.p[i]%=min;
        if(vec_A.p[i] != 0) not_all_zero = true;
      };
  } while (not_all_zero);
  g=min;
  vec_A.free();
}; /* Vector_Gcd */


/* normalize a vector in such a way that gcd(vec)=1 */
void Vector_Normalize(int* vec, int size, int dimension)
{
  int gcd, i, *p;
  
  Vector_Gcd(vec, dimension, gcd);
  if (gcd>=2) {
    p=vec;
    for (i=0;i<size;i++)
      *p++/=gcd;
  }
}; /* Vector_Normalize */


/* computes r3 as combination of r1 and r2 in such a way that r3[k]=0 */
void combine(int* r1, int* r2, int* r3, int k, int nbcolumns, int dim)
{
  int aux;
  int aux2;
  int r1j, r2j;

  Gcd(r1[k], r2[k], &aux);
  int a1= *(r1+k)/aux;
  int a2= *(r2+k)/aux;

  // this appears to be the site of most overflow problems
  //   for (int j=0;j<nbcolumns;j++) r3[j]=mymult(a2,r1[j])-mymult(a1,r2[j]);
  mult_call_nr = 1;
  for (int j=0;j<nbcolumns;j++) {
    if (overflow_avoid_mode) {
      int cum, this_a1, this_a2;
      this_a1 = a1;
      this_a2 = a2;
      Gcd(r1[j],r2[j],&aux2);
      if (aux2 == 0) aux2 = 1;
      r1j = r1[j]/aux2;
      r2j = r2[j]/aux2;
      cum= aux2;

      if (overflow_avoid_mode >=2) {
	// try with r1j and a1, and also r2j and a2
	Gcd(r1j,this_a1,&aux2);
	if (aux2 == 0) aux2 = 1;
	r1j /= aux2;
	this_a1 /= aux2;
	cum *= aux2;
// if (aux2 >=2) cout << "1) Saved factor of " << aux2 << "\n";

	Gcd(r2j,this_a2,&aux2);
	if (aux2 == 0) aux2 = 1;
	r2j /= aux2;
	this_a2 /= aux2;
	cum *= aux2;
// if (aux2 >=2) cout << "2) Saved factor of " << aux2 << "\n";
      }

	
      r3[j]=mymult((mymult(this_a2,r1j)-mymult(this_a1,r2j)),cum);
    }
    else {
      r1j = r1[j];
      r2j = r2[j];
      r3[j]=(mymult(a2,r1j)-mymult(a1,r2j));
    }
  }
  Vector_Normalize(r3, nbcolumns, dim);
}; /* combine */

int chernikova(matrix* bid, matrix* uni,  int dimension, int& nbbidray, int& nbray,
               vector* ineq, int start)
{
/*
     change: start is the index of the first constraint not already taken into 
     account.  (so, the standard algorithm always consider start=0)
  */

  matrix ray, bidray;

  int nbcolumns, index_non_zero, nbcommonconstraints;
  int i, j, k, l, m, *p, bound, inf_bound, equal_bound, sup_bound, redundant;

  bidray=*bid;
  ray=*uni;
  vector inequality=*ineq;           
  nbcolumns=bidray.nbcolumns; /* width of both matrices uni and bid          */

  vector commonzero(nbcolumns);

  for (k=dimension+start;k<nbcolumns;k++) { /* for each constraint           */
    /* finds if exists, a bidirectional ray which does not saturate the      */
    /* current constraint                                                    */
    index_non_zero=-1;
    for (i=0;i<nbbidray;i++)
      if (bidray.p[i][k]!=0) {
        index_non_zero=i;
        break;
      }
    if (index_non_zero!=-1) { /* if a such ray exists                        */
      /* discards index_non_zero bidirectional ray                           */
      nbbidray--;
      if (nbbidray!=index_non_zero) {
        p=bidray.p[nbbidray];
        bidray.p[nbbidray]=bidray.p[index_non_zero];
        bidray.p[index_non_zero]= p;
      }
      /* Computes the new lineality space                                    */
      for (i=0;i<nbbidray;i++)
        if (bidray.p[i][k]!=0)
          combine(bidray.p[i], bidray.p[nbbidray], bidray.p[i], k, 
                  nbcolumns, dimension);
      /* add the positive part of index_non_zero bidirectional ray to the    */
      /* set of unidirectional rays                                          */
      if (nbray==MAXRAYS)
        return 1;
      if (bidray.p[nbbidray][k]<0)
        for (j=0;j<nbcolumns;j++)
          ray.p[nbray][j]=-bidray.p[nbbidray][j];
      else
        for (j=0;j<nbcolumns;j++)
          ray.p[nbray][j]=bidray.p[nbbidray][j];
      /* Computes the new pointed cone                                       */
      for (i=0;i<nbray;i++)
        if (ray.p[i][k]!=0)
          combine(ray.p[i], ray.p[nbray], ray.p[i], k, nbcolumns, dimension);
      if (inequality.p[k])
        nbray++;
    }
    else { /* all bidirectional rays belongs to the current supporting       */
           /* hyperplane                                                     */
      /* sort rays : 0 <= i < equal_bound : saturates the constraint         */
      /*           : equal_bound <= i < sup_bound : verifies the constraint  */
      /*           : sup_bound <= i < bound : does not verify                */
      /* Just a three classes linear sorting algorithm                       */
      equal_bound=0;
      sup_bound=0;
      inf_bound=nbray;
      while (inf_bound>sup_bound) {
        if (ray.p[sup_bound][k]==0) {
          EXCH(ray, equal_bound, sup_bound, p);
          equal_bound++;
          sup_bound++;
        }
        else if (ray.p[sup_bound][k]<0) {
          inf_bound--;
          EXCH(ray, inf_bound, sup_bound, p);
       }
        else sup_bound++;
      }

      /* Computes only the new pointed cone by combining opposite            */
      /* (with respect to the new constraint) rays                           */
      bound=nbray;
      for (i=equal_bound;i<sup_bound;i++) /* select a positive ray           */
        for(j=sup_bound;j<bound;j++) {    /* select a negative ray           */
          /* computes the set of common saturated constraints                */
          nbcommonconstraints=0;
          for (l=dimension;l<k;l++)
            if ((ray.p[i][l]==0) && (ray.p[j][l]==0)) {
              commonzero.p[nbcommonconstraints]=l;
              nbcommonconstraints++;
            }
          if (nbcommonconstraints+nbbidray>=dimension-2) {
            /* if its a candidate for extremality                            */
            /* check whether a ray m saturates the same set of constraints   */
            redundant=false;
            for (m=0;m<bound;m++)
              if ((m!=i) && (m!=j)) { /* don't consider the current rays     */
                for (l=0;l<nbcommonconstraints;l++)
                  if (ray.p[m][commonzero.p[l]]!=0) break;
                if (l==nbcommonconstraints) {
                  /* the combination of ray i and j will generate a          */
                  /* non extremal ray so discards it                         */
                  redundant=true;
                  break;
                }
              }
            if (!redundant) { /* add the new ray to the set                  */
              if (nbray==MAXRAYS)
                return 1;
              /* computes the new ray                                        */
              combine(ray.p[j], ray.p[i], ray.p[nbray], k, nbcolumns, 
                      dimension);
              nbray++;
            }
          }
        }
      /* Eliminates all non extremal rays                                    */
      if (inequality.p[k])
        j=sup_bound;
      else
        j=equal_bound;
      i=nbray;
      while ((j<bound)&&(i>bound)) {
        i--;
        EXCH(ray, i, j, p);
        j++;
      }
      if (j==bound) nbray=i;
      else nbray=j;
    }
  }
  commonzero.free();
  
  /* set the size of resulting matrices bid and uni                          */
  bid->nbrows=nbbidray;
  uni->nbrows=nbray;
  return 0;
} /* chernikova_body */

matrix* matrix:: share()
/* copy with shared array */
{
   return new matrix(nbrows,nbcolumns,p_init);
}; /* share_matrix */

matrix* matrix::copy()
/* copy with copied array */
{
  matrix* c = new matrix(nbrows, nbcolumns);
  for(int i=0;i< nbrows;i++)
    for(int j=0; j< nbcolumns; j++)
       c->p[i][j] = p[i][j];
  c->nonstrict = nonstrict;
  return c;
}; /* copy_matrix*/

bool ori_translate(matrix& F, matrix& C, matrix* FR , matrix* CR, 
                   vector* trans_vec, bool dual_given)
/* 
   translates  the frame F  to the origin, result in FR ; 
   if 'dual-given' is true, the same translation 
   is performed on the constraint system C , result in CR
*/
{
  int nbrows, nbcolumns, b0,b, i,j,k;
  int prod;
  bool nul;
  matrix fr, cr;
  *FR = matrix();
  *CR = matrix();

  nbrows = F.nbrows-1;
  nbcolumns = F.nbcolumns;
  /* is  the origin already a vertex? */
  nul = true;
  for(i=0; i<nbrows; i++) if(F.p[i][nbcolumns-1]){
       nul = true;
       for(j=1; j<nbcolumns-1; j++) if(F.p[i][j]){
         nul = false; break;
       }
     if(nul) break;
  };
  if(nbrows==0 || nul){
  /* the frame contains the origin, no translation */
      *FR = matrix(F); 
      if(dual_given) *CR = matrix(C);
      *trans_vec = vector(nbcolumns);
      for (j=1;j<nbcolumns-1;j++) trans_vec->p[j] = 0;
      trans_vec->p[0]=1;  trans_vec->p[j] = 1;
      return(false);
  }

  /* the frame doesn't contain the origin: translation */
  fr = matrix(F.nbrows,nbcolumns);
  if(dual_given) cr = matrix(C.nbrows,C.nbcolumns);

  /* find the first vertex V, which will be the translation vector */
  for(i=0; (i<nbrows && (F.p[i][0]==0 || F.p[i][nbcolumns-1]==0));i++)
     for(j=0;j<nbcolumns;j++) fr.p[i][j] = F.p[i][j];
  if (i==nbrows) {
         printf("Error: no vertices\n"); exit(2);
  };

  *trans_vec = vector(nbcolumns);
  for(j=0;j<nbcolumns;j++)trans_vec->p[j] = F.p[i][j];
  /* the origin is now a vertex */
  for(j=1;j<nbcolumns-1;j++) fr.p[i][j] =0;
  fr.p[i][0]=1;  fr.p[i][j]=1;

  /* translate all other vertices but the last */
  b0 = F.p[i][nbcolumns-1];
  for(k=i+1; k<nbrows; k++){
     if(F.p[k][0] != 0 && F.p[k][nbcolumns-1] != 0){
        b=F.p[k][nbcolumns-1];

 mult_call_nr = 2;

        for(j=1;j<nbcolumns-1;j++) 
              fr.p[k][j] = mymult(b0,F.p[k][j]) - mymult(b,trans_vec->p[j]);
        fr.p[k][0] = 1; fr.p[k][j] = mymult(b0,b); 
     }
     else /* rays and lines are copied */
     for(j=0;j<nbcolumns;j++)  fr.p[k][j] =F.p[k][j];
  }
  /* the last row */
  for(j=1;j<nbcolumns-1;j++) fr.p[k][j] = 0;
  fr.p[k][0]=1;  fr.p[k][j]=1;
  *FR = fr;

  if(dual_given){ /* translate the constraints: B <- B - A*trans_vec */
     for(i=0;i<C.nbrows;i++){
          prod = 0; cr.p[i][0] = C.p[i][0];
          for(j=1;j<nbcolumns-1;j++) {
//              prod += C.p[i][j] * trans_vec->p[j];

 mult_call_nr  = 3;

              prod += mymult(C.p[i][j],trans_vec->p[j]);
              cr.p[i][j] = C.p[i][j];
          };
          cr.p[i][j] = C.p[i][j] + prod;
      };
     *CR = cr;
   };
  return(true);
}; /* ori_translate */

void ori_restore(matrix *mat, vector *trans_vec)
/* translate a matrix representing a constraint system according to a vector */
{
  int nbrows, nbcolumns, prod, not_nul_row, b0, b0_norm, g, i,j,k;
  nbrows = mat->nbrows;
  nbcolumns = mat->nbcolumns;

vec_norm_call_nr = 0;

  Vector_Normalize(trans_vec->p,trans_vec->size,trans_vec->size);
  b0 = trans_vec->p[nbcolumns-1];
  if(b0==0){
     printf("bad translation vector\n"); exit(0);
  }
  k=0;
  for(i=0;i<nbrows-1;i++){
     prod=0; not_nul_row=0; mat->p[k][0]=mat->p[i][0];
     for(j=1;j<nbcolumns-1;j++){
       if(mat->p[i][j]){
          not_nul_row = 1;
//          prod += mat->p[i][j] * trans_vec->p[j]; 

 mult_call_nr = 4;

          prod += mymult(mat->p[i][j],trans_vec->p[j]); 
       }
     }
     if(not_nul_row){
        Gcd(b0,prod,&g); b0_norm = b0/g;

 mult_call_nr = 5;

        for(j=1;j<nbcolumns-1;j++)
//           mat->p[k][j]= b0_norm * mat->p[i][j];
           mat->p[k][j]= mymult(b0_norm, mat->p[i][j]);
//        mat->p[k][j] = b0_norm * mat->p[i][j] - prod/g;


 mult_call_nr = 6;

        mat->p[k][j] = mymult(b0_norm, mat->p[i][j]) - prod/g;
        k++;
     }
  }
  if(k!=i){
     mat->p[k][0]=1; 
     for(j=1;j<nbcolumns-1;j++) mat->p[k][j] = 0;
     mat->p[k][j] = 1;
     mat->nbrows--;
  }

}; /* ori_restore */

void transform(matrix& m, matrix *n, matrix *result, bool constraints_wanted, int known)
/*
   Computes, according to  'constraints_wanted'
       - either the frame of the polyhedron defined by constraints 'm', already knowing 
         the frame 'n' corresponding to the first 'known' constraints of 'm'.
       - or the constraint system of the polyhedron whose frame is 'm', already knowing
         the constraint system 'n' of the first 'known' elements of 'm'.

 Assumes that the result matrix is "empty" (dataless) at time of call.
 Thus no need to free it.
*/
{  
   int nbconstraints, dimension, nbcolumns, nbcolumns2, nbrays, nbbidrays, i, j, k;
   bool need_restoration, dual_given;
   int prod;
   matrix wm, wn;
   vector trans_vec;
   bool debug = false;



   nbconstraints=m.nbrows;
   nbcolumns=m.nbcolumns;
   dimension=nbcolumns-1;
   nbcolumns2=dimension+nbconstraints;
   dual_given = (known>0);

   if(debug){
      printf("Transform\n"); m.print(); if(dual_given)n->print();
   };

   need_restoration= false;
   if(constraints_wanted) /* constraints desired: translation to origin */
        need_restoration= ori_translate(m, *n, &wm, &wn, &trans_vec, dual_given);
   else{
       wm = m; 
       if(dual_given) wn = *n;
   };

   if(debug){
       printf("Actual systems\n"); wm.print(); if(dual_given)wn.print();
   };

   matrix ray(MAXRAYS, nbcolumns2);
   matrix bidray(dimension, nbcolumns2);
   

   if(dual_given){ // A part of the result is already known
      // initialization of the frame
      nbrays=0; nbbidrays=0;
      for(i=0;i<wn.nbrows-1;i++){
         if(wn.p[i][0]){ // vertex or ray
            for(j=1;j<wn.nbcolumns;j++) ray.p[nbrays][j-1] = wn.p[i][j];
            nbrays++;
         }
         else { // line
            for(j=1;j<wn.nbcolumns;j++) bidray.p[nbbidrays][j-1] = wn.p[i][j];
            nbbidrays++;
         }
      }
      bidray.nbrows = nbbidrays; 

      // initialization of saturations with respect to the matrix
      for(i=0;i<nbrays;i++){
        for(k=0;k<nbconstraints;k++){
           prod = 0;

 mult_call_nr = 7;

           for(j=1;j<nbcolumns;j++) prod += mymult(ray.p[i][j-1],wm.p[k][j]);
           ray.p[i][dimension+k] = prod;
        }
      }
      for(i=0;i<nbbidrays;i++){
        for(k=0;k<nbconstraints;k++){
           prod = 0;
//           for(j=1;j<nbcolumns;j++) prod += bidray.p[i][j-1]*wm.p[k][j];

 mult_call_nr = 8;

           for(j=1;j<nbcolumns;j++) 
	     prod += mymult(bidray.p[i][j-1],wm.p[k][j]);
           bidray.p[i][dimension+k] = prod;
        }
      }
   }
   else { // initialization to the frame of the whole space
       /* set the identity matrix as bidirectional rays */
       for (i=0;i<dimension;i++) {
         for (j=0;j<dimension;j++) bidray.p[i][j]=0;
         bidray.p[i][i]=1;
       };

       nbbidrays = dimension; nbrays = 0;

       for (i=0;i<nbconstraints;i++)
         for (j=1;j<nbcolumns;j++)
           bidray.p[j-1][dimension+i]=wm.p[i][j];
   }


  vector inequality(nbcolumns2);
  for (i=0;i<nbconstraints;i++) inequality.p[dimension+i]=wm.p[i][0];

  /*** Print the saturations 
  printf("Saturations: %d rays, %d lines\n", nbrays, nbbidrays);
  printf("     Lines\n"); bidray.print();
  printf("     Rays\n"); ray.print();
  printf("    Inequalities\n"); inequality.print();
  */

  if (chernikova(&bidray, &ray, dimension, nbbidrays, nbrays, &inequality, known)!=0) {
     printf("not enough rays available\n");
     exit(2);
  }

  if(debug){
     printf("CHERNIKOVA\n"); bidray.print(); ray.print();
     printf("Dimensions:%2d %2d\n", dimension, nbcolumns2);
  }

/*
  // Minimize the initial system
  // find all the equations
  int nbeq=0;
  for(k=dimension; k<nbcolumns2-1;k++){
    if(inequality.p[k]){
      for(j=0;j<ray.nbrows;j++){if(ray.p[j][k]!=0) break;}
      if(j==ray.nbrows){inequality.p[k]=false; nbeq++;}
    }
    else nbeq++;
  }
  vector redundant(nbcolumns2);
  for(j=dimension; j<nbcolumns2;j++) redundant.p[j]=0;
  vector zero(ray.nbrows); int nbsat;
  for(i=nbcolumns2-2;i>=dimension; i--){
     if(inequality.p[i]){ // for each inequality
        nbsat=0; for(j=0;j<ray.nbrows;j++){if(ray.p[j][i]==0)zero.p[nbsat++]=j;}
        // must be saturated at least by (dimension-2) elements
        if(nbsat+bidray.nbrows < (dimension-nbeq-2)){redundant.p[i]=true; continue; }
        // is there another inequality saturated by the same rays
        for(k=dimension; k<nbcolumns2-1;k++){ 
           if(k!=i && inequality.p[k] && !redundant.p[k]){
              for(j=0;j<nbsat;j++) {
                  if(ray.p[zero.p[j]][k]!=0) break;
              }
              if(j==nbsat) {
                  redundant.p[i]=true; break;
              }
            }
        }
     }
  }
  k=0;
  for(i=0;i<m.nbrows;i++){
    if(!redundant.p[dimension+i]) m.p[k++] = m.p[i];
  }
  m.nbrows = k; 

  if(debug){
     printf("Redundancies\n");
     for(j=dimension; j<=nbcolumns2-1;j++) printf("%2d",redundant.p[j]); printf("\n");
     printf("Simplified constraints\n"); m.print();
  }
*/
  /* changed HWT, 3/21/95, to free result first */
   // 9/95
   // no longer needed, assume result pts to an "empty" (dataless) matrix
   // was result->free();

  *result = matrix(bidray.nbrows+ray.nbrows+1,dimension+1);

  for (i=0;i<bidray.nbrows;i++){
     result->p[i][0] = 0;
     for (j=0;j<dimension;j++) result->p[i][j+1] = bidray.p[i][j];
  };
  for (k=0; k<ray.nbrows;k++){
     result->p[i][0] = 1;
     for (j=0; j<dimension;j++) result->p[i][j+1] = ray.p[k][j];
     i++;
  }
  // the last row
  result->p[i][0] = 1;
  for (j=1;j<dimension;j++) result->p[i][j] = 0;
  result->p[i][j] = 1;

  if(need_restoration){ /* restoration of the origin */
         ori_restore(result, &trans_vec);
         wm.free(); if(dual_given) wn.free();
  };

  if(debug){
      printf("Result\n"); m.print(); result->print();
  };


  if(constraints_wanted) trans_vec.free(); 

  // HWT 5/5/95. block containing construction of 
  // redundant and free are commented out
  // redundant.free(); 
  // zero.free();

  inequality.free();
  bidray.free(); ray.free();
}; /* transform */



