#include "compose.H"

// #define DEBUG_SLOPES
// #define DEBUG_MATRIX_STORE
// #define DEBUG_COMPOSE    -- misc print outs
// #define DEBUG_INVARIANT
// #define DEBUG_TRANS


int nr_freed = 0;


// find the array position of a key string in an array of strings of size leng
// return -1 if not found
int
string_index(char *key_str, char **str_table, int leng) {
  int i=0;
  bool found = false;

  while ((i<leng) && !found) {
    if (! (found = (strcmp(key_str,str_table[i])==0))  ) {
      i++;
    }
  }
  if (!found) {
    return (-1);
  }
  else {
    return(i);
  }
}
int
find_reach_index(int i, int j, 
		 automaton *aut1, automaton *aut2, 
		 int *is_reached) {
  int prod_index = i * aut2->nr_locations + j;
  return( is_reached[prod_index] );
}
  


    
// add the component info for *autp as the comp_nr'th component
// assume all components before this one have had their info installed
// already
// technically, comp_nr is redundant, since this info should be contained
// in the nr_components field.
// copy new names
void
automaton::add_comp_info(int comp_nr, automaton *autp) {
  int i;
  if (comp_nr != nr_components) {
    cout << "Error: call to add_comp_info() with mismatch in nr components\n";
    cout << flush;
    exit(1);
  }
  if (comp_nr==0) {
    comp_inf_p = (component_info *) malloc (sizeof(component_info));
  }
  else {
    comp_inf_p = 
      (component_info *) realloc (comp_inf_p,
				 (comp_nr+1)*sizeof(component_info));
  }
  comp_inf_p[comp_nr].automaton_name = str_preserve(autp->name);
  comp_inf_p[comp_nr].nr_locations = autp->nr_locations;
  comp_inf_p[comp_nr].locn_names = 
     (char **) malloc (autp->nr_locations * sizeof(char *));
  for (i=0;i<autp->nr_locations;i++) {
    comp_inf_p[comp_nr].locn_names[i] = str_preserve(autp->locations[i].name);
  }
  nr_components++;
}


void
automaton::copy_comp_info(automaton *aut1) {
  int i,j;
  nr_components = aut1->nr_components;
  if (nr_components == 0) {
    add_comp_info(0,aut1);
  }
  else {
    comp_inf_p = 
      (component_info *) malloc (nr_components * sizeof(component_info ));
    for (i=0;i<nr_components;i++) {
      comp_inf_p[i].automaton_name = 
	str_preserve(aut1->comp_inf_p[i].automaton_name);
      comp_inf_p[i].nr_locations = aut1->comp_inf_p[i].nr_locations;
      comp_inf_p[i].locn_names = 
	(char **) malloc (comp_inf_p[i].nr_locations * sizeof(char *));
      for (j=0;j<comp_inf_p[i].nr_locations;j++) {
	comp_inf_p[i].locn_names[j] = 
	  str_preserve(aut1->comp_inf_p[i].locn_names[j]);
      }
    }
  }
}

// make the component info correct (used in forming product with *autp)
void 
automaton::set_up_comp_inf(automaton *aut1,automaton *aut2){
  // if this is the first component, then put in data for the first process
  copy_comp_info(aut1);
  add_comp_info(nr_components,aut2);
}





// given ptrs to two automaton, set up nr_sync_labels, and the actual labels
void
automaton::set_up_alphabet(automaton *aut1, automaton *aut2) {
// simple 2 pass algorithm
  int i,j;
  int nr_events;   // number of unique event names, assuming unique within
                   // each automaton already

  // first determine how many unique events there are
  nr_events = aut1->nr_sync_labels;

  for (i=0;i<aut2->nr_sync_labels;i++) {
    if (string_index(aut2->sync_label_names[i],
		     aut1->sync_label_names,
		     aut1->nr_sync_labels) == -1) {
      nr_events++;
    }
  }
  nr_sync_labels = nr_events;
  // allocate space
  sync_label_names = (char **) malloc (nr_events * sizeof(char *));
  for (i=0;i<aut1->nr_sync_labels;i++) {
    sync_label_names[i] = str_preserve(aut1->sync_label_names[i]);
  }
  j = i;
  for (i=0;i<aut2->nr_sync_labels;i++) {
    if (string_index(aut2->sync_label_names[i],
		     aut1->sync_label_names,
		     aut1->nr_sync_labels) == -1) {      
     sync_label_names[j] = str_preserve(aut2->sync_label_names[i]);
     j++;
   }
  }
#ifdef DEBUG_COMPOSE
  cout << "In composed alphabet ... \n" << flush;
  cout << "Nr sync labels = " << nr_sync_labels << "\nSync labels are:\n";
  for (i=0;i<nr_sync_labels;i++) {
    cout << "\t" << sync_label_names[i] << "\n"<< flush;
  }
#endif
  
}


// return ptr to a new string, which is concat of s1 and s2
char *
str_concat(char *s1, char *s2) {
  char *p, *q, *cats;
  cats = (char *) malloc ((strlen(s1) + strlen(s2) + 1) * sizeof(char));
  p = cats;
  q = s1;
  while (*p++ = *q++) ;
  // go back to null terminator
  p--;
  q = s2;
  while (*p++ = *q++) ;
  return ( cats );
}

char *product_name(char *s1, char *s2) {
  char *s1dot = str_concat(s1,".");
  char *prod_str  = str_concat(s1dot,s2);
  free(s1dot);
  return(prod_str);
}

// return lcm of the global list of slope vectors
// 
int
lcm_of(slope_vector *sl,int nr_sl) {
  int i;
  int temp = 1;
  for (i=0;i<nr_sl;i++) {
    temp = lcm(temp, sl[i].p[var_count+1]);
  }
  return(temp);
}


// return ptr to array of ptrs to of slope-vectors, all scaled by newlcm
slope_vector *
get_scaled_slopes(slope_vector *sl, int nr_sl, int newlcm) {
  slope_vector *temp = 
    (slope_vector *) malloc (nr_sl * sizeof(slope_vector));
  int i,j;
  int extra_factor;
  for (i=0;i<nr_sl;i++) {
    temp[i] = slope_vector(sl[i]);
    // the current scale-up factor recorded in the last entry of the vector
    extra_factor = newlcm / sl[i].p[var_count+1];
    for (j=0;j<var_count;j++) {
      temp[i].p[j+1] = extra_factor * sl[i].p[j+1];
    }
    temp[i].p[var_count+1] = newlcm;
  }

  return ( temp);
}

void
free_scaled_slopes(slope_vector *svp, int nr_sl) {
  int i;
  for (i=0;i<nr_sl;i++) {
    svp[i].free();
  }
  global_free(svp);
}


int
is_slope_for_undecl_var(slope_vector *sv) {
// 
// check whether it is a line
//
  return(sv->p[0] == 0);
}

void
show_intpp(int **app, int nr_locs) {
// assume array is of size nr_locs, var_count
  int i,j;
  for (i=0;i<nr_locs;i++) {
    cout << "For locn " << i << "\n";
    for (j=0;j<var_count;j++) {
      cout << i << " " << j << " " << app[i][j] << "\n";
    }
    cout << "\n";
  }
  cout << "+++++++++++++++++++\n";
}


// use separate procedures for min/max/decl for now

// used the scaled slopes sslope to determine min/max values

// 
int **
get_min_slopes(automaton *aut,slope_vector *sslopes) {
  int i,j,k;
  slope_vector *this_slope;
  int *started = (int *) malloc (var_count * sizeof(int));
  int **mins = (int **) malloc (aut->nr_locations *sizeof(int *));
  for (i=0;i<aut->nr_locations;i++) {
    mins[i] = (int *) malloc (var_count * sizeof (int));
    clear_array(mins[i], var_count);
    clear_array(started,var_count);
    for (j=0;j<aut->locations[i].nr_slopes;j++) {
      // ptr subtraction gives correct # of items between ptrs
#ifdef DEBUG_SLOPES
      trace_out("Value of diff btw slopes is ", 
		aut->locations[i].post_slopes[j]-
			    &aut->slopes[0]);
#endif
      this_slope = &sslopes[aut->locations[i].post_slopes[j]-
			    &aut->slopes[0]];
      if (!is_slope_for_undecl_var(this_slope)) {
	for (k=0;k<var_count;k++) {
	  if (started[k]) {
	    if (this_slope->p[k+1]<mins[i][k]) {
	      mins[i][k] = this_slope->p[k+1];
	    }
	  }
	  else {
	    mins[i][k] = this_slope->p[k+1];
	    started[k] = true;
	  }
	}
      }
    }
  }
#ifdef DEBUG_SLOPES
cout << "Min values \n";
show_intpp(mins,aut->nr_locations);
#endif
  global_free(started);
  return (mins);
}



// 
int **
get_undecls(automaton *aut) {
  int i,j,k;
  slope_vector *this_slope;
  int **undecl = (int **) malloc (aut->nr_locations *sizeof(int *));
  for (i=0;i<aut->nr_locations;i++) {
    undecl[i] = (int *) malloc (var_count * sizeof (int));
    // assume all declared
    clear_array(undecl[i], var_count);
    for (j=0;j<aut->locations[i].nr_slopes;j++) {
      // ptr subtraction gives correct # of items between ptrs
      this_slope = aut->locations[i].post_slopes[j];
      if (is_slope_for_undecl_var(this_slope)) {
	for (k=0;k<var_count;k++) {
	  if (this_slope->p[k+1] != 0) {
	    undecl[i][k] = true;
	  }
	}
      }
    }
  }
#ifdef DEBUG_SLOPES
cout << "undecl? values \n";
show_intpp(undecl,aut->nr_locations);
#endif
  return (undecl);
}

// 
int **
get_max_slopes(automaton *aut,slope_vector *sslopes) {
  int i,j,k;
  slope_vector *this_slope;
  int *started = (int *) malloc (var_count * sizeof(int));
  int **maxs = (int **) malloc (aut->nr_locations *sizeof(int *));
  for (i=0;i<aut->nr_locations;i++) {
    maxs[i] = (int *) malloc (var_count * sizeof (int));
    clear_array(maxs[i], var_count);
    clear_array(started,var_count);
    for (j=0;j<aut->locations[i].nr_slopes;j++) {
      // ptr subtraction gives correct # of items between ptrs
      this_slope = &sslopes[aut->locations[i].post_slopes[j]-
			    &aut->slopes[0]];
      if (!is_slope_for_undecl_var(this_slope)) {
	for (k=0;k<var_count;k++) {
	  if (started[k]) {
	    if (this_slope->p[k+1]>maxs[i][k]) {
	      maxs[i][k] = this_slope->p[k+1];
	    }
	  }
	  else {
	    maxs[i][k] = this_slope->p[k+1];
	    started[k] = true;
	  }
	}
      }
    }
  }
#ifdef DEBUG_SLOPES
cout << "Max values \n";
show_intpp(maxs,aut->nr_locations);
#endif
  global_free(started);
  return (maxs);
}




void
set_min_max(automaton *aut1, automaton *aut2, 
	    slope_vector *scaled_slopes_1, 
	    slope_vector *scaled_slopes_2,
	    int ***  slope_1_min, int ***slope_1_max, int ***undecl_1, 
	     int *** slope_2_min, int ***slope_2_max, int ***undecl_2) {
// for int **, 1st index is for location, 2nd for variable (column)
// the *** is the address of int **, since we're changing its value.

  *slope_1_min = get_min_slopes(aut1,scaled_slopes_1);
  *slope_1_max = get_max_slopes(aut1,scaled_slopes_1);
  *undecl_1 = get_undecls(aut1);
  *slope_2_min = get_min_slopes(aut2,scaled_slopes_2);
  *slope_2_max = get_max_slopes(aut2,scaled_slopes_2);
  *undecl_2 = get_undecls(aut2);
}



// add the slope vector to global slopes of automaton.
// also set up pointer to it from locations[ind]. no do it for vecindex
//    since the actual pt'ed to slopes will change on realloc.
void
automaton::install_vector(slope_vector sv, int ind, int ** vecindex) {
  // add space then copy vector
  if (nr_slopes>0) {
    slopes = (slope_vector *) realloc (slopes, 
				      (nr_slopes+1) * sizeof(slope_vector));
  }
  else {
    // allocate memory for the first vector
    slopes = (slope_vector *) malloc (sizeof(slope_vector));
  }
  slopes[nr_slopes] = sv;
  nr_slopes++;
  if (locations[ind].nr_slopes > 0) {
    vecindex[ind] = (int *) realloc (vecindex[ind],
				 (locations[ind].nr_slopes+1)* sizeof (int));
  }
  else {
    // set up space for the first item
    vecindex[ind] = (int *) malloc (sizeof(int));
  }

#ifdef DEBUG_SLOPES
trace_out("ind is ", ind);
trace_out("insert position is", locations[ind].nr_slopes);
trace_out("about to install posn as ", nr_slopes-1);
#endif

  vecindex[ind][locations[ind].nr_slopes] = nr_slopes-1;
  locations[ind].nr_slopes++;


}


// recursive procedure, based on var_nr
// adds product slopes, incrementally building up vec
//    and splitting whenever mins and maxs differ
// does not worry about the slopes for undeclared variables
// aut is the product automaton we're trying to build.
// 5/11/95: Mem. cut leaks for recursive call with new vectors
void
add_p_s(int var_nr, slope_vector vec, int index, int **vecindex,
	automaton *aut,
	int *min1,int *max1,int *und1,int *min2,int *max2,int *und2) {
#ifdef DEBUG_SLOPES
 cout << "IN addps with varnr " << var_nr << "\n";
 vec.print();
#endif
// cout << "KKKKKKKKKKKK\n";
   if (var_nr >= var_count) {
     // bottom of recursion
     // add the vector to the list of slopes and put in a ptr to it
     vec.p[var_count+1] = aut->slope_lcm;
     // make it a ray, since derived from ray slopes, not undecl var slopes
     vec.p[0] = 1;
#ifdef DEBUG_SLOPES
     vec.print();
cout << "About to install at bottom of recursion\n" << flush;
#endif
     aut->install_vector(vec,index,vecindex);
   }
   else {
     if (und1[var_nr]) {
       if (und2[var_nr]) {
	 add_p_s(var_nr+1,vec,  index,vecindex,aut, min1,max1,und1,min2,max2,und2);
       }
       else { // declared in P1
	 slope_vector vv(vec);
	 vv.p[var_nr+1] = min2[var_nr];
	 add_p_s(var_nr+1,vv, index,vecindex,aut, min1,max1,und1,min2,max2,und2);
	 if (min2[var_nr] != max2[var_nr]) {
	   slope_vector vv2(vec);
	   vv2.p[var_nr+1] = max2[var_nr];
	   add_p_s(var_nr+1,vv2, index,vecindex,aut, min1,max1,und1,min2,max2,und2);
	 }
	   // 5/11/95: free the vector coming in, since new vector used 
	   // on way out (recursive calls)
	 vec.free();
       }
     }
     else { 
       // var_nr is declared in aut 1's loc
       if (und2[var_nr]) {
	 slope_vector vv(vec);
	 vv.p[var_nr+1] = min1[var_nr];
	 add_p_s(var_nr+1,vv, index,vecindex,aut, min1,max1,und1,min2,max2,und2);
	 if (min1[var_nr] != max1[var_nr]) {
	   slope_vector vv2(vec);
	   vv2.p[var_nr+1] = max1[var_nr];
	   add_p_s(var_nr+1,vv2, index,vecindex,aut, min1,max1,und1,min2,max2,und2);
	 }
	 vec.free();
       } 
       else {
	 // both declared
	 int minval = max_of(min1[var_nr], min2[var_nr]);
	 int maxval = min_of(max1[var_nr], max2[var_nr]);
#ifdef DEBUG_SLOPES
	 cout << "min is " << minval << ", and max is " << maxval << "\n";
#endif
	 if (minval <= maxval) {
	   slope_vector vv(vec);
	   vv.p[var_nr+1] = minval;
	   add_p_s(var_nr+1,vv, index,vecindex,aut, min1,max1,und1,min2,max2,und2);
	   if (minval != maxval) {
	     slope_vector vv2(vec);
	     vv2.p[var_nr+1] = maxval;
	     add_p_s(var_nr+1,vv2, index,vecindex,aut, min1,max1,und1,min2,max2,und2);
	   }
	   vec.free();
	 }
	 // else {empty} so don't continue
       }
     }
   }
}

// return vector for unbounded increase in dimension of var_nr
// use the constant 1 as lcm for now.
slope_vector
undecl_vec( int var_nr) {
  int i;
  slope_vector vec(var_count+2);
  // make it strongly sloped
  vec.p[0] = 0;
  for (i=0;i<var_count;i++) {
    vec.p[i+1] = 0;
  }
  vec.p[var_nr+1] = vec.p[var_count+1] = 1;
  return(vec);
}


// for the paired state (i,j) of aut1 and aut2, add the product vectors for
// the slopes.  may also (yes for now) add to the automaton's global list
// and set the ptr from the locations[ (i,j) ].post_slopes too
// assumes lcm field has been set (for the slopes)
void 
automaton::set_product_slopes(int i,int j, automaton *aut2, 
			 int **slope_1_min,int **slope_1_max,int **undecl_1,
			 int **slope_2_min,int **slope_2_max,int **undecl_2,
			 int **vecindex) {
  int k;
  slope_vector vec(var_count+2);
  for (k=0;k<var_count+1;k++) {
#ifdef DEBUG_SLOPES
    if (vec.p[k] != 0) {
      cout << "NON-zero after creating slope vec!! \n" << flush;
    }
#endif
    vec.p[k] = 0;
  }
  int index = i*aut2->nr_locations + j;
  int *min1 = slope_1_min[i];
  int *max1 = slope_1_max[i];
  int *und1 = undecl_1[i];
  int *min2 = slope_2_min[j];
  int *max2 = slope_2_max[j];
  int *und2 = undecl_2[j];

                           // says which posn in global slope array is used
                           // for the post_slopes of this location.

  // 1st param is var number
  add_p_s(0,vec,index,vecindex,this,min1,max1,und1,min2,max2,und2);

  // now add slopes for undeclared variables.
  // ...
  for (k=0;k<var_count;k++) {
    if (und1[k] && und2[k]) {
      slope_vector vec = undecl_vec(k);
      vec.p[k+1] = vec.p[var_count+1] = slope_lcm;
#ifdef DEBUG_SLOPES
     cout << "About to install vector for undeclared var " << k << "\n" << flush;
#endif
      install_vector(vec,index,vecindex);
    }
  }

}


// all the above routines can be ditched in new implementation of slopes
// now use slope matrices or rather polys.
// done within set_up_locations ...

   
bool
syntactic_matrix_eq(matrix *m1, matrix *m2) {
  int i,j;
  int nbr = m1->nbrows;
  int nbc = m1->nbcolumns;
  if ((nbr != m2->nbrows) || (nbc != m2->nbcolumns)) {
    return(false);
  }
  for (i=0;i<nbr-1;i++) {
    for (j=0;j<nbc;j++) {
      if (m1->p[i][j] != m2->p[i][j]) {
	return(false);
      }
    }
  }
  return(true);
}



// put the new matrix in the matrix store (don't use a fresh copy of data)
// set (index) ptr for matindex for posn in matrix store
// index is the position of the location.
// the ptr to indices for this location (index) has size elts so far
   // 5/9/95: changed to take addr of ptr to matrix.  If matrix known to appear
   // already then free matrix, and set ptr to previous occurrence of matrix
   // in the store.
   // space_flag indicates how prudent to be about saving space
   // space_flag = 0 => install matrix regardless
   //            = 1 => search beyond invariant matrices
   //            = 2 => search over all matrices
   // search uses syntactic equivalence of matrices, no fiddling with
   //   minimizing matrices, though this would be better.
void 
automaton::install_matrix(matrix **mpaddr,int index, int **matindex, int size,
			  int space_flag) {
  // set up bounds for search
  // currently only use with space_flag == 1, since invariant matrices
  //    handled separately
  int lowindex, highindex;
  bool found = false;
  int pos;
  switch (space_flag) {
  case 0: 
    lowindex = 0; highindex = -1;
    break;
  case 1: 
    lowindex = nr_locations + 1;
    highindex = nr_matrices_in_store-1;
    break;
  case 2:
    lowindex = 0;
    highindex = nr_matrices_in_store-1;
    break;
  }
  // set found to true if matrix already exists in store, or -1 if not found
  // move low index to high index.
  // only check matrix if it's "original"
  // if found, set pos to position found at.
  while (!found && (lowindex<=highindex)) {
    found = syntactic_matrix_eq(&(matrix_store[lowindex]),*mpaddr);
    if (found) {
      pos = lowindex;
    }
    else {
      lowindex++;
    }
  }

  if (found) {
    (*mpaddr)->free();
    *mpaddr = &(matrix_store[pos]);
  //  cerr << "Freeing matrix\n";
	nr_freed++;
    matindex[index][size] = pos;	
    return;
  }

  // otherwise must install the matrix
  if (nr_matrices_in_store == 0) {
    matrix_store = (matrix *) malloc (sizeof(matrix));
  }
  else {
    matrix_store =
      (matrix *) realloc(matrix_store, 
			 (nr_matrices_in_store+1) * sizeof(matrix));
  }
  matrix_store[nr_matrices_in_store] = matrix(**mpaddr);
  if (size == 0) {
    matindex[index] = (int *) malloc (sizeof(int));
  }
  else {
    matindex[index] = 
      (int *) realloc(matindex[index], (size+1) * sizeof(int));
  }
  matindex[index][size] = nr_matrices_in_store;
  nr_matrices_in_store++;
}

// return ptr to free matrix which is intersection of m1 and m2 in
// constraint forms
// assume we are working with an nplus1_form matrix, with eps_pos = 
// nbcolumns-2-eps_adj
matrix 
*matrix_intersection(matrix *m1, matrix *m2, int eps_adj = 0) {
  matrix *resultm;
  matrix *cm1 = m1->copy();   // copy of the matrix and data

  // polyhdron constructor copies ptrs only, not data
  polyhedron pp(*cm1,true);
  pp.intersection(*m2);
  resultm =  pp.constraints()->copy();
  pp.free();

  resultm->nonstrict = 
	!(resultm->nplus1_is_strict(resultm->nbcolumns-2-eps_adj));
  return(resultm);
}

// return ptr to free matrix which is intersection of m1 and m2 in
// constraint forms
void
matrix_test(matrix *m1, matrix *m2) {
  matrix *resultm;
//  matrix *cm1 = m1->copy();   // copy of the matrix and data

  // polyhdron constructor copies ptrs only, not data
//  polyhedron pp(*cm1,true);
//  resultm =  pp.constraints();
//  pp.free();
  resultm = m2;
  matrix F(2,3);
  // test frame
  transform(*m1, ((matrix*) 0), &F, false, 0);
  F.free();
}

// set up ptr to initial location, and matrix for init region
// OK for this matrix to be sitting in free space, unattached 
// to other matrices.
void 
automaton::set_up_init_info(automaton *aut1, automaton *aut2) {
  init_location_p = 
    &locations[aut1->init_location_p->locn_id * aut2->nr_locations +
	       aut2->init_location_p->locn_id];
  init_region_p = matrix_intersection(aut1->init_region_p,
					      aut2->init_region_p);
  
}

// set up ptr to initial location, and matrix for init region
// OK for this matrix to be sitting in free space, unattached 
// to other matrices.
void 
automaton::set_up_init_info_reach(automaton *aut1, automaton *aut2, 
				    int *is_reached) {
  int index = find_reach_index(aut1->init_location_p->locn_id, 
			       aut2->init_location_p->locn_id,
			       aut1,
			       aut2,
			       is_reached);
  if (index == -1) {
    cerr << "Error: Init locn not reached!\n";
    exit (1);
  }

  init_location_p = 
    &locations[ index];
  init_region_p = matrix_intersection(aut1->init_region_p,
					      aut2->init_region_p);
  
}

// create a matrix for the invariant of product location
// with location i in aut1 and loc j in aut2.
// store matrix in matrix_store, and put indexptr to it in matindex
// by convention we put the invariant matrices all at the beginning
// of matrix_store!
void 
automaton::set_up_product_invariant(int i, int j, 
				    automaton *aut1, automaton *aut2) {
  bool found = false;
  int pos;
  int p;
  int index = i*aut2->nr_locations+j;
#ifdef DEBUG_INVARIANT
  cout << "setting up prod invar for " << i << " " << j << "\n" << flush;
  trace_out("index is ", index);
  aut1->locations[i].invariant->print();
  aut2->locations[j].invariant->print();
#endif

  matrix *mp = matrix_intersection(aut1->locations[i].invariant,
				   aut2->locations[j].invariant);

  // set found to true if matrix already exists in store, and false otherwise
  // move low index to high index.
  // only check matrix if it's "original"
  // if found, set pos to position found at.
  p = 1;
  while (!found && (p<=nr_locations)) {
    found = syntactic_matrix_eq(&(matrix_store[p]),mp);
    if (found) {
      pos = p;
    }
    else {
      p++;
    }
  }
  if (found) {
    mp->free();
    delete mp;
    // copy ptrs only, not data
    matrix_store[index] = matrix_store[pos];
  }
  else {
    is_orig_matrix[index] = true;
    matrix_store[index] = *mp;
  }

#ifdef DEBUG_MATRIX_STORE
  cout << "Matrix for product location:\n" << flush;
  matrix_store[index].print();
  cout << "Matrices for previous locations\n" << flush;
  {
    int ijk;
    for (ijk=0;ijk<index;ijk++) {
      matrix_store[ijk].print();
    }
  }
  cout << " ~~~~~~~~~~~~\n" << flush;
#endif
}



// create a matrix for the invariant of product location
// with location i in aut1 and loc j in aut2.
// store matrix in matrix_store, and put indexptr to it in matindex
// by convention we put the invariant matrices all at the beginning
// of matrix_store!
void 
automaton::set_up_reach_product_invariant(int i, int j, 
				    automaton *aut1, automaton *aut2, 
					  int index) {
  bool found = false;
  int pos;
  int p;
  int prod_index = i*aut2->nr_locations+j;
#ifdef DEBUG_INVARIANT
  cout << "setting up prod invar for " << i << " " << j << "\n" << flush;
  trace_out("index is ", index);
  aut1->locations[i].invariant->print();
  aut2->locations[j].invariant->print();
#endif

  matrix *mp = matrix_intersection(aut1->locations[i].invariant,
				   aut2->locations[j].invariant);

  // set found to true if matrix already exists in store, and false otherwise
  // move low index to high index.
  // only check matrix if it's "original"
  // if found, set pos to position found at.
// 11/1/95: why did p start at 1 and go to nr_locations before????
  p = 0;
  while (!found && (p<nr_locations)) {
    found = syntactic_matrix_eq(&(matrix_store[p]),mp);
    if (found) {
      pos = p;
    }
    else {
      p++;
    }
  }
  if (found) {
    mp->free();
    delete mp;
    // copy ptrs only, not data
    matrix_store[index] = matrix_store[pos];
  }
  else {
    is_orig_matrix[index] = true;
    matrix_store[index] = *mp;
  }
}


bool
has_sync_label(transition tr) {
  return ( tr.sync_label != NULL );
}

bool
label_in_alphabet(char *keystr, char **name_array, int size) {
// 
  int i = 0;
  bool found = false;
  while (!found && (i<size)) {
    found = (strcmp(keystr,name_array[i]) == 0);
    if (!found) {
      i++;
    }
  }
  return (found);
}

// add one more transition to end of *trp, already of size *size
// don't add anything to the space yet
// increment *size
void
increase_sizes(transition ** trp, int **actualguardindex, int **actualassindex,
	       int index, int *size) {
  if (*size == 0) {

    *trp = (transition *) malloc (sizeof(transition));
    actualguardindex[index] = (int *) malloc (sizeof(int));
    actualassindex[index] = (int *) malloc (sizeof(int));
  }
  else {
    *trp = (transition *) realloc(*trp, (*size+1) * sizeof(transition));
    actualguardindex[index] = (int *) realloc ((int *)actualguardindex[index],
					       (*size+1) * sizeof(int));
    actualassindex[index] = (int *) realloc ((int *)actualassindex[index],
					     (*size+1) * sizeof(int));
  }
  (*size)++;
#ifdef DEBUG_COMPOSE
  trace_out("Incre for index ", index);
  trace_out("Incre to size ", (*size));
  { int kk;
    for (kk=0;kk<(*size)-1;kk++) {
      // trp[kk].print();
      trace_out("In inc. size gd idx = ", actualguardindex[index][kk]);
      trace_out("In inc. size as idx = ", actualassindex[index][kk]);
    }
  }
#endif
}

// NOTE: this routine deals with unsynchronized transitions
// add a post_trans, number tnr of aut1's location locn_nr_1
// allocate space for it in post_trans of this automaton's locn index
// store guard and assignment matrices in matrix_store, and update
// info in the matindices
// assume that locn2 stays the same.
// which = 1 or 2  for source  of unsynch (single) trans
// if = 2, then only diff is recompute the destindex 
// because the locn_nr_1 have been reversed already on call.
void
automaton::add_trans(int which,
		     int tnr, automaton *aut, int locn_nr_1, int locn_nr_2,
		     int size2, int index,
		  int **actualguardindex, int **actualassindex) {
  int dest_index;
  int use_locn_nr;
  int post_trans_index;
#ifdef DEBUG_TRANS
  trace_out("In ADD TRANS with *************** ", which);
#endif
  if (which ==1) {
    dest_index = (aut->locations[locn_nr_1].post_trans[tnr].dest->locn_id) *
      size2 + locn_nr_2;
    use_locn_nr = locn_nr_1;
  }
  else {
    dest_index = (locn_nr_1 * size2 + 
		  aut->locations[locn_nr_2].post_trans[tnr].dest->locn_id);
#ifdef DEBUG_TRANS
trace_out("$$$$ inside 2nd $$$ with dest locn id ",
	  aut->locations[locn_nr_2].post_trans[tnr].dest->locn_id);
trace_out(aut->locations[locn_nr_2].post_trans[tnr].dest->name, 0);
#endif
    use_locn_nr = locn_nr_2;

  }
#ifdef DEBUG_TRANS
trace_out("dest index = ", dest_index);
#endif
  post_trans_index = locations[index].nr_post_trans;
  increase_sizes(&locations[index].post_trans, 
		 actualguardindex, actualassindex, 
		 index,
		 &locations[index].nr_post_trans);
  // watch out. nr_post_trans already been incremented by increase_sizes
  // store matrices for guard and assignment
  matrix *en_mat_p = aut->locations[use_locn_nr].post_trans[tnr].enabling_constr->copy();
  install_matrix(&en_mat_p,
		 index, actualguardindex, post_trans_index,1);
  matrix *ass_mat_p = aut->locations[use_locn_nr].post_trans[tnr].
		   ass.ass_constraints->copy();
  install_matrix(&ass_mat_p,
		 index, actualassindex, post_trans_index,1);
#ifdef DEBUG_TRANS
trace_out("installed matrices for guard and constraints. index ", index);
trace_out("now number post trans is ", locations[index].nr_post_trans);
trace_out("new gd index value is ", actualguardindex[index][locations[index].nr_post_trans-1]);
trace_out("new ass index value is ", actualassindex[index][locations[index].nr_post_trans-1]);

  {int pp;
   for (pp=0;pp<locations[index].nr_post_trans;pp++) {
trace_out("new gd index value is ", actualguardindex[index][pp]);
trace_out("new ass index value is ", actualassindex[index][pp]);
}
}     
#endif

  // i.e. put in dummy matrices for now, fix later in adjust_matrix_ptrs()
  locations[index].post_trans[post_trans_index] =
    transition(&locations[index], &locations[dest_index], 
	       aut->locations[use_locn_nr].post_trans[tnr].sync_label,
	       (matrix *) NULL, 
	       aut->locations[use_locn_nr].post_trans[tnr].is_asap_trans,
	       new assignment(
			aut->locations[use_locn_nr].post_trans[tnr].ass.nr_vars,
			aut->locations[use_locn_nr].post_trans[tnr].ass.indices,
			(matrix *) NULL ) );

  
}


// NOTE: this routine deals with unsynchronized transitions
// add a post_trans, number tnr of aut1's location locn_nr_1
// allocate space for it in post_trans of this automaton's locn index
// store guard and assignment matrices in matrix_store, and update
// info in the matindices
// assume that locn2 stays the same.
// which = 1 or 2  for source  of unsynch (single) trans
// if = 2, then only diff is recompute the destindex 
// because the locn_nr_1 have been reversed already on call.
void
automaton::add_reach_trans(int which,
			   int tnr, automaton *aut1, automaton *aut2, 
			   int locn_nr_1, int locn_nr_2,
			   int size2, int index,
			   int **actualguardindex, int **actualassindex,
			   int *is_reached) {
  int dest_index;
  int use_locn_nr;
  int post_trans_index;
  automaton* aut;
#ifdef DEBUG_TRANS
  trace_out("In ADD TRANS with *************** ", which);
#endif
  if (which ==1) {
    dest_index = 
      find_reach_index(aut1->locations[locn_nr_1].post_trans[tnr].dest->locn_id,
		       locn_nr_2,
		       aut1, aut2,
		       is_reached);
    aut = aut1;
    use_locn_nr = locn_nr_1;
  }
  else {
    dest_index = 
      find_reach_index(locn_nr_1,
		      aut2->locations[locn_nr_2].post_trans[tnr].dest->locn_id,
		       aut1,aut2,
		       is_reached);
    aut = aut2;
    use_locn_nr = locn_nr_2;
  }
#ifdef DEBUG_TRANS
trace_out("dest index = ", dest_index);
#endif
  post_trans_index = locations[index].nr_post_trans;
  increase_sizes(&locations[index].post_trans, 
		 actualguardindex, actualassindex, 
		 index,
		 &locations[index].nr_post_trans);
  // watch out. nr_post_trans already been incremented by increase_sizes
  // store matrices for guard and assignment
  matrix *en_mat_p = aut->locations[use_locn_nr].post_trans[tnr].enabling_constr->copy();
  install_matrix(&en_mat_p,
		 index, actualguardindex, post_trans_index,1);
  matrix *ass_mat_p = aut->locations[use_locn_nr].post_trans[tnr].
		   ass.ass_constraints->copy();
  install_matrix(&ass_mat_p,
		 index, actualassindex, post_trans_index,1);
#ifdef DEBUG_TRANS
trace_out("installed matrices for guard and constraints. index ", index);
trace_out("now number post trans is ", locations[index].nr_post_trans);
trace_out("new gd index value is ", actualguardindex[index][locations[index].nr_post_trans-1]);
trace_out("new ass index value is ", actualassindex[index][locations[index].nr_post_trans-1]);

  {int pp;
   for (pp=0;pp<locations[index].nr_post_trans;pp++) {
trace_out("new gd index value is ", actualguardindex[index][pp]);
trace_out("new ass index value is ", actualassindex[index][pp]);
}
}     
#endif

  // i.e. put in dummy matrices for now, fix later in adjust_matrix_ptrs()
  locations[index].post_trans[post_trans_index] =
    transition(&locations[index], &locations[dest_index], 
	       aut->locations[use_locn_nr].post_trans[tnr].sync_label,
	       (matrix *) NULL, 
	       aut->locations[use_locn_nr].post_trans[tnr].is_asap_trans,
	       new assignment(
			aut->locations[use_locn_nr].post_trans[tnr].ass.nr_vars,
			aut->locations[use_locn_nr].post_trans[tnr].ass.indices,
			(matrix *) NULL ) );

  
}


//  given two assignments, find matrix and indices, etc for the product assgnmt
// put it in prod_ass
void
get_product_assignment_info(assignment *prod_ass,
			      assignment ass1, assignment ass2) {
  // figure out how many columns
  // allocate space for the matrix, and index array
  // set up ptrs from second automaton's reassigned vars to the first's 
  //          and also whether they are already included in the first at all
  // copy over the info from the first assignment matrix
  // copy over info from the second.
  int nr_ass_vars_1 = ass1.nr_vars;
  int nr_ass_vars_2 = ass2.nr_vars;
  int nr_cols;
  int i,j;
  // 1 per ass_vars in ass2
  // ass_posn stores whether the ass's ass vars are in ass 1 already
  // if they're not, then new_ass_pos stores where they will be put
  int *ass_posn = (int *) malloc (nr_ass_vars_2 * sizeof(int));
  int *new_ass_posn = (int *)  malloc (nr_ass_vars_2 * sizeof(int));
  // find out how many of the second assignment's ass vars are not already
  // assigned in assignmt1, also set up ptrs to which posn they are in if new.
  // also set the number of vars in the prod ass.
  prod_ass->nr_vars = ass1.nr_vars;
  for (i=0;i<nr_ass_vars_2;i++) {
    ass_posn[i] = 
      pos_in_int_array(ass2.indices[i],ass1.indices,nr_ass_vars_1);
    // a new assgt var will be unfound, and hence return value -1
    if (ass_posn[i] == -1) {
      new_ass_posn[i] = prod_ass->nr_vars;
      prod_ass->nr_vars++;
    }
    else {
      new_ass_posn[i] = ass_posn[i];
    }
  }
  // now allocate the space
  nr_cols = var_count + prod_ass->nr_vars+2;
  // assumes there is an assignment matrix there, even if no assignment done
  prod_ass->ass_constraints = 
    new matrix(ass1.ass_constraints->nbrows + ass2.ass_constraints->nbrows -1,
	       nr_cols);
  prod_ass->indices = (int *) malloc (prod_ass->nr_vars * sizeof(int));
  
  // set the matrix
  // copy ass1's matrix (stretching out the constant column)
  // make this a procedure?
  for (i=0;i<ass1.ass_constraints->nbrows-1;i++) {
    for (j=0;j<ass1.ass_constraints->nbcolumns-1;j++) {
      prod_ass->ass_constraints->p[i][j] = ass1.ass_constraints->p[i][j];
    }
    for (j=ass1.ass_constraints->nbcolumns-1;j<nr_cols-1;j++) {
      prod_ass->ass_constraints->p[i][j] = 0;
    }
    prod_ass->ass_constraints->p[i][nr_cols-1] = 
      ass1.ass_constraints->p[i][ass1.ass_constraints->nbcolumns-1];
  }
  // copy info from ass2's matrix
  for (i=0;i<ass2.ass_constraints->nbrows-1;i++) {
    // copy all the columns for var_count unprimed vars, and eq column
    for (j=0;j<var_count+1;j++) {
      prod_ass->ass_constraints->p[ass1.ass_constraints->nbrows-1+i][j] = ass2.ass_constraints->p[i][j];
    }
    // blank out all other columns, and set constant column
    for (j=var_count+1;j<nr_cols-1;j++) {
      prod_ass->ass_constraints->p[ass1.ass_constraints->nbrows-1+i][j] = 0;
    }
    prod_ass->ass_constraints->p[ass1.ass_constraints->nbrows-1+i][nr_cols-1] = 
      ass2.ass_constraints->p[i][ass2.ass_constraints->nbcolumns-1];
    // now copy over the relevant numbers for assigned vars
    for (j=0;j<ass2.nr_vars;j++) {
      prod_ass->ass_constraints->p[ass1.ass_constraints->nbrows-1+i][var_count+1+new_ass_posn[j]] =
	ass2.ass_constraints->p[i][var_count+1+j];
    }
  }

  // add the final row
  prod_ass->ass_constraints->p[prod_ass->ass_constraints->nbrows-1][0] = 1;
  for (j=1;j<prod_ass->ass_constraints->nbcolumns-1;j++) {
    prod_ass->ass_constraints->p[prod_ass->ass_constraints->nbrows-1][j] = 0;
  }
  prod_ass->ass_constraints->p[prod_ass->ass_constraints->nbrows-1]
                              [nr_cols-1] = 1;

  // set the indices
  // copy from first ass
  for (i=0;i<ass1.nr_vars;i++) {
    prod_ass->indices[i] = ass1.indices[i];
  }
  // and now from the second
  for (i=0;i<ass2.nr_vars;i++) {
    // actually the test here is unnecessary
    if (ass_posn[i] == -1) {
      prod_ass->indices[new_ass_posn[i]] = ass2.indices[i];
    }
  }
  prod_ass->ass_constraints->nonstrict = 
    !prod_ass->ass_constraints->
	nplus1_is_strict(prod_ass->ass_constraints->nbcolumns-2-
			 prod_ass->nr_vars);
  free(ass_posn);
  free(new_ass_posn);

}



// add a synchronized transition.  using
// trans tnr1 of aut1's id1 locn, and trans tnr2 of aut2's id2 locn
// store indexptrs in actualguardindex and actualassindex
void
automaton::add_sync_trans(int tnr1, int tnr2,automaton *aut1, automaton *aut2,
	       int id1, int id2, 
	       int **actualguardindex,
	       int **actualassindex) {
  int dest_index = (aut1->locations[id1].post_trans[tnr1].dest->locn_id) *
      aut2->nr_locations + aut2->locations[id2].post_trans[tnr2].dest->locn_id;
  int index = id1*aut2->nr_locations + id2;
  int post_trans_index = locations[index].nr_post_trans;

#ifdef DEBUG_TRANS
  cout << "In ADD SYNC TRANS \n";
  cout << "Add Sync trans UNIMPLEMENTED\n";
#endif
  increase_sizes(&locations[index].post_trans, 
		 actualguardindex, actualassindex, 
		 index,
		 &locations[index].nr_post_trans);
  // watch out. nr_post_trans already been incremented by increase_sizes
  // store matrices for guard and assignment
  assignment prod_ass;
  // get the matrix for the guard for the product event
  matrix *mgd = 
    matrix_intersection(aut1->locations[id1].post_trans[tnr1].enabling_constr,
			aut2->locations[id2].post_trans[tnr2].enabling_constr);
  get_product_assignment_info(&prod_ass,
			      aut1->locations[id1].post_trans[tnr1].ass,
			      aut2->locations[id2].post_trans[tnr2].ass);
  install_matrix(&mgd,
		 index, actualguardindex, post_trans_index,1);
  install_matrix(&(prod_ass.ass_constraints),
		 index, actualassindex, post_trans_index,1);

  // i.e. put in dummy matrices for now, fix later in adjust_matrix_ptrs()
  locations[index].post_trans[post_trans_index] =
    transition(&locations[index], &locations[dest_index], 
	       aut1->locations[id1].post_trans[tnr1].sync_label,
	       (matrix *) NULL, 
	       aut1->locations[id1].post_trans[tnr1].is_asap_trans ||
	       aut2->locations[id2].post_trans[tnr2].is_asap_trans,
	       &prod_ass );

}



// add a synchronized transition.  using
// trans tnr1 of aut1's id1 locn, and trans tnr2 of aut2's id2 locn
// store indexptrs in actualguardindex and actualassindex
void
automaton::add_sync_reach_trans(int tnr1, int tnr2,
				automaton *aut1, automaton *aut2,
				int id1, int id2, 
				int **actualguardindex,
				int **actualassindex,
				int index,
				int *is_reached) {
  int dest_index = 
    find_reach_index(aut1->locations[id1].post_trans[tnr1].dest->locn_id,
	       aut2->locations[id2].post_trans[tnr2].dest->locn_id, 
	       aut1, aut2, is_reached);
  if (dest_index == -1) {
    cerr << "Error in product, unreachable locn is reached!\n";
    exit (1);
  }
//  int prod_index = id1*aut2->nr_locations + id2;
  int post_trans_index = locations[index].nr_post_trans;

#ifdef DEBUG_TRANS
  cout << "In ADD SYNC TRANS \n";
  cout << "Add Sync trans UNIMPLEMENTED\n";
#endif
  increase_sizes(&locations[index].post_trans, 
		 actualguardindex, actualassindex, 
		 index,
		 &locations[index].nr_post_trans);
  // watch out. nr_post_trans already been incremented by increase_sizes
  // store matrices for guard and assignment
  assignment prod_ass;
  // get the matrix for the guard for the product event
  matrix *mgd = 
    matrix_intersection(aut1->locations[id1].post_trans[tnr1].enabling_constr,
			aut2->locations[id2].post_trans[tnr2].enabling_constr);
  get_product_assignment_info(&prod_ass,
			      aut1->locations[id1].post_trans[tnr1].ass,
			      aut2->locations[id2].post_trans[tnr2].ass);
  install_matrix(&mgd,
		 index, actualguardindex, post_trans_index,1);
  install_matrix(&(prod_ass.ass_constraints),
		 index, actualassindex, post_trans_index,1);

  // i.e. put in dummy matrices for now, fix later in adjust_matrix_ptrs()
  locations[index].post_trans[post_trans_index] =
    transition(&locations[index], &locations[dest_index], 
	       aut1->locations[id1].post_trans[tnr1].sync_label,
	       (matrix *) NULL, 
	       aut1->locations[id1].post_trans[tnr1].is_asap_trans ||
	       aut2->locations[id2].post_trans[tnr2].is_asap_trans,
	       &prod_ass );

}



// set up transitions, except use indexptrs to the matrices (ref to 
//  matrix_store index posn).
void 
automaton::set_up_trans(int id1, int id2, automaton *aut1, automaton *aut2,
			int **actualguardindex, int **actualassindex) {
// for each transition in aut1, test to see if it has to be synchronized with
// transitions in aut2.  if yes, produce a transition for each match
// else produce a single transition
// for each transition in aut 2, produce a single transition if  no sync label
  int i,j;
  bool add_single_trans;
  int index = id1*aut2->nr_locations + id2;

#ifdef DEBUG_TRANS
  trace_out ("Checking for loc ",id1);
  trace_out(" with loc ", id2);
#endif
  for (i=0;i<aut1->locations[id1].nr_post_trans;i++) {
#ifdef DEBUG_TRANS
    trace_out("\t first aut, trans ", i);
#endif
    add_single_trans = false;
    if (has_sync_label(aut1->locations[id1].post_trans[i])) {
      char *label = aut1->locations[id1].post_trans[i].sync_label;
      if (label_in_alphabet(label,aut2->sync_label_names,aut2->nr_sync_labels))
	{
	for (j = 0;j<aut2->locations[id2].nr_post_trans;j++) {
#ifdef DEBUG_TRANS
	  trace_out(" \t\t second aut trans ",j);
#endif
	  if (has_sync_label(aut2->locations[id2].post_trans[j])) {
	    if (strcmp(label,
		       aut2->locations[id2].post_trans[j].sync_label) == 0) {


#ifdef DEBUG_TRANS
	      cout << "Sync on this trans\n";
#endif
	      add_sync_trans(i,j,aut1,aut2,id1,id2,
			     actualguardindex,actualassindex);
	      
	    }
	  }
	}
      }
      else {
	// label does not appear in other aut's alphabet
	add_single_trans = true;
      }
    }
    else {
      // no sync label.  just add single transition
      add_single_trans = true;
    }
    if (add_single_trans) {
      // add the transition
      add_trans(1,i,aut1,id1,id2,aut2->nr_locations,index,
		actualguardindex,actualassindex);
    }
  }
  for (i=0;i<aut2->locations[id2].nr_post_trans;i++) {
    // add if either no sync label, or a sync label not in A1's alphabet
    if (!has_sync_label(aut2->locations[id2].post_trans[i]) ||
	(has_sync_label(aut2->locations[id2].post_trans[i]) &&
	 !label_in_alphabet(aut2->locations[id2].post_trans[i].sync_label,
			   aut1->sync_label_names,aut1->nr_sync_labels))) {
      // add single transition
      add_trans(2,i,aut2,id1,id2,aut2->nr_locations,index,
		actualguardindex,actualassindex);
      
    }
  }
#ifdef DEBUG_TRANS
  cout << "Done with locs " << id1 << " and " << id2 << "\n" << flush;
#endif
}


// set up transitions, except use indexptrs to the matrices (ref to 
//  matrix_store index posn).
// given index # for this reachable prod. location as last input arg.
void 
automaton::set_up_reach_trans(int id1, int id2, 
			      automaton *aut1, automaton *aut2,
			      int **actualguardindex, int **actualassindex,
			      int index, int *is_reached) {
// for each transition in aut1, test to see if it has to be synchronized with
// transitions in aut2.  if yes, produce a transition for each match
// else produce a single transition
// for each transition in aut 2, produce a single transition if  no sync label
  int i,j;
  bool add_single_trans;
  int prod_index = id1*aut2->nr_locations + id2;

#ifdef DEBUG_TRANS
  trace_out ("Checking for loc ",id1);
  trace_out(" with loc ", id2);
#endif
  for (i=0;i<aut1->locations[id1].nr_post_trans;i++) {
#ifdef DEBUG_TRANS
    trace_out("\t first aut, trans ", i);
#endif
    add_single_trans = false;
    if (has_sync_label(aut1->locations[id1].post_trans[i])) {
      char *label = aut1->locations[id1].post_trans[i].sync_label;
      if (label_in_alphabet(label,aut2->sync_label_names,aut2->nr_sync_labels))
	{
	for (j = 0;j<aut2->locations[id2].nr_post_trans;j++) {
#ifdef DEBUG_TRANS
	  trace_out(" \t\t second aut trans ",j);
#endif
	  if (has_sync_label(aut2->locations[id2].post_trans[j])) {
	    if (strcmp(label,
		       aut2->locations[id2].post_trans[j].sync_label) == 0) {


#ifdef DEBUG_TRANS
	      cout << "Sync on this trans\n";
#endif
	      add_sync_reach_trans(i,j,aut1,aut2,id1,id2,
				   actualguardindex,actualassindex, 
				   index, is_reached);
	      
	    }
	  }
	}
      }
      else {
	// label does not appear in other aut's alphabet
	add_single_trans = true;
      }
    }
    else {
      // no sync label.  just add single transition
      add_single_trans = true;
    }
    if (add_single_trans) {
      // add the transition
      add_reach_trans(1,i,aut1,aut2,id1,id2,aut2->nr_locations,index,
		actualguardindex,actualassindex,is_reached);
    }
  }
  for (i=0;i<aut2->locations[id2].nr_post_trans;i++) {
    // add if either no sync label, or a sync label not in A1's alphabet
    if (!has_sync_label(aut2->locations[id2].post_trans[i]) ||
	(has_sync_label(aut2->locations[id2].post_trans[i]) &&
	 !label_in_alphabet(aut2->locations[id2].post_trans[i].sync_label,
			   aut1->sync_label_names,aut1->nr_sync_labels))) {
      // add single transition
      add_reach_trans(2,i,aut1,aut2,id1,id2,aut2->nr_locations,index,
		actualguardindex,actualassindex,is_reached);
      
    }
  }
#ifdef DEBUG_TRANS
  cout << "Done with locs " << id1 << " and " << id2 << "\n" << flush;
#endif
}



// put in correct ptrs to matrices, based on the index info
void 
automaton::adjust_ptrs_to_matrices(int loc1, int loc2, 
				    automaton *aut2,
				   int **actualguardindex, 
				   int **actualassindex) {
  int i;
  int index = loc1*aut2->nr_locations + loc2;
  locations[index].invariant = &(matrix_store[index]);
#ifdef DEBUG_COMPOSE
  cout << "In adjust ptrs to matrices " << loc1 << " and " << loc2 << "\n" << flush;

  trace_out("Index is ", index); 
trace_out("Number matr is store ", nr_matrices_in_store);
locations[index].invariant->print();
  trace_out("Number of post_trans = ",locations[index].nr_post_trans);
#endif

  for (i=0;i<locations[index].nr_post_trans;i++) {
#ifdef DEBUG_COMPOSE
    trace_out("\t actualguardindex : ", actualguardindex[index][i]);
#endif
    locations[index].post_trans[i].enabling_constr = 
      &matrix_store[actualguardindex[index][i]];
#ifdef DEBUG_COMPOSE
    trace_out("\t actualassindex : ", actualassindex[index][i]);
#endif
    locations[index].post_trans[i].ass.ass_constraints =
      &matrix_store[actualassindex[index][i]];
  }
}




// put in correct ptrs to matrices, based on the index info
void 
automaton::adjust_ptrs_to_matrices_reach(int loc1, int loc2, 
					 automaton *aut2,
					 int **actualguardindex, 
					 int **actualassindex,
					 int index) {
  int i;
  int prod_index = loc1*aut2->nr_locations + loc2;
  locations[index].invariant = &(matrix_store[index]);
#ifdef DEBUG_COMPOSE
  cout << "In adjust ptrs to matrices " << loc1 << " and " << loc2 << "\n" << flush;

  trace_out("Index is ", index); 
trace_out("Number matr is store ", nr_matrices_in_store);
locations[index].invariant->print();
  trace_out("Number of post_trans = ",locations[index].nr_post_trans);
#endif

  for (i=0;i<locations[index].nr_post_trans;i++) {
#ifdef DEBUG_COMPOSE
    trace_out("\t actualguardindex : ", actualguardindex[index][i]);
#endif
    locations[index].post_trans[i].enabling_constr = 
      &matrix_store[actualguardindex[index][i]];
#ifdef DEBUG_COMPOSE
    trace_out("\t actualassindex : ", actualassindex[index][i]);
#endif
    locations[index].post_trans[i].ass.ass_constraints =
      &matrix_store[actualassindex[index][i]];
  }
}


void
automaton::free_actual_indexes(int **actualguardindex, int **actualassindex) {
   int i;
   for (i=0;i<nr_locations;i++) {
     global_free(actualguardindex[i]);
     global_free(actualassindex[i]);
   }
   global_free(actualguardindex);
   global_free(actualassindex);
}


void
free_slope_info(int **slope_inf, int nr_locs) {
   int i;
   for (i=0;i<nr_locs;i++) {
     // this frees an array of size varcount of integers
     global_free(slope_inf[i]);
   }
   global_free(slope_inf);
}




// set up info regarding the locations, except for the transitions
// but we don't want to destroy the current automaton's locations do we?
// also set up vector arrays, and matrix arrays
void
automaton::set_up_locns(automaton *aut1, automaton *aut2) {
// need to 
// 1) determine how many locations
// 2) allocate space
// 3) set names
// 4) set slopes for each
// 5) set inv matrix
// 6) set backwards indices for easy access to component locations ???
//
  int i,j;
  int **slope_1_min = NULL, **slope_1_max=NULL;
  int **slope_2_min = NULL, **slope_2_max = NULL;
  int ** undecl_1 = NULL, **undecl_2 = NULL;
//  int new_lcm;
  int nr_locns_1, nr_locns_2;
//  slope_vector * scaled_slopes_1, *scaled_slopes_2;
  int **actualguardindex, ** actualassindex;  // first index is locn, second
                                              // is transition
  int **actualvecindex;  // index of positions of vectors for each locn
  nr_locns_1 = aut1->nr_locations;
  nr_locns_2 = aut2->nr_locations;
  nr_locations = nr_locns_1*nr_locns_2;
  locations = (location *) malloc (nr_locations * sizeof(location));

  // set up empty global list of slopes
  nr_slopes = 0;
  slopes = (slope_vector *) NULL;
  actualvecindex = (int **) malloc (nr_locations * sizeof(int *));
  clear_ptr_array(actualvecindex,nr_locations);

  // set up the scaled slopes (so everyone's on the same scale)  
  // 
/*
  new_lcm = lcm(1,lcm_of(aut1->slopes,aut1->nr_slopes));
  new_lcm = lcm(new_lcm,lcm_of(aut2->slopes,aut2->nr_slopes));
  scaled_slopes_1 = get_scaled_slopes(aut1->slopes,aut1->nr_slopes, new_lcm);
  scaled_slopes_2 = get_scaled_slopes(aut2->slopes,aut2->nr_slopes,new_lcm);

  slope_lcm = new_lcm;
  set_min_max(aut1,aut2,scaled_slopes_1, scaled_slopes_2,
	      &slope_1_min, &slope_1_max, &undecl_1, 
	      &slope_2_min, &slope_2_max, &undecl_2);

  free_scaled_slopes(scaled_slopes_1,aut1->nr_slopes);
  free_scaled_slopes(scaled_slopes_2,aut2->nr_slopes);
*/
 
  // don't need invmatrix index since just use first nr_locations spots
  //  invmatrixindex = (int *) malloc (nr_locations * sizeof(int));
  //  clear_array(invmatrixindex,nr_locations);
  // allocate space for the first entries, all of which are inv matrices


nr_freed = 0;

  // set up each individual location
  // for each aut1 location
  for (i=0;i<nr_locns_1;i++) {
    // for each aut 2 location
    for (j=0;j<nr_locns_2;j++) {
      // place most defaults in position, and ...
      int index = i*nr_locns_2+j;
      locations[index] = location();
      locations[index].locn_id = index;
      // set up names
      locations[index].name = product_name(aut1->locations[i].name,
					       aut2->locations[j].name);

      // set up product post slope matrix
      locations[index].post_slopes_matrix = 
	aut1->locations[i].post_slopes_matrix->copy()->
          intersection_no_simp(*(aut2->locations[j].post_slopes_matrix->constraints()));

	locations[index].post_slopes_matrix->fix_strict();
	locations[index].post_slopes_matrix->constraints()->nonstrict =  
	  !locations[index].post_slopes_matrix->strict;   

if   (! locations[index].post_slopes_matrix->constraints()->nonstrict) {
  cout << "Strict rate conditions !\n";
  cout << "locations[index].post_slopes_matrix->eps_pos = " <<
    locations[index].post_slopes_matrix->eps_pos << "\n";
}
/*
      // set up slopes
      // test out realloc()
      // slopes now include the lcm used to scale them up.
      // this is so we can recover the actual rates for comparison with
      // actual rates in other automata, i.e. relative to a time unit.
      // lcm stored in constant column for rate vectors.
      //
      // thus we now have to choose the min and max for each var taking into
      // a/c whether declared.
      // for each var, if not declared, take min and max of other
      //                               but if other not declared, skip
      //               else (* declared *) then
      //                    if other not declared, take min and max
      //                           else pairwise min and max
      //               if min = max then just use value, don't split.
      // add undecl vectors at end.

      set_product_slopes(i,j, aut2, 
			 slope_1_min,slope_1_max,undecl_1,
			 slope_2_min,slope_2_max,undecl_2,
			 actualvecindex);
*/
#ifdef DEBUG_SLOPES
      cout << "The slopes we get for " << i << " and " << j << "\n";
      trace_out("There are this many of them : ",locations[index].nr_slopes);
#endif

    }
  }

/*
 for (i=0;i<nr_locns_1;i++) {
   for (j=0;j<nr_locns_2;j++) {
     int index = i*aut2->nr_locations+j;
     // copy ptrs instead of using temporary vecindex for refering to slopes
     locations[index].post_slopes = 
       (slope_vector **) malloc (locations[index].nr_slopes * 
				 sizeof(slope_vector *));
     for (k=0;k<locations[index].nr_slopes;k++) {
       locations[index].post_slopes[k] = &slopes[actualvecindex[index][k]];
     }
#ifdef DEBUG_SLOPES
      cout << "The slopes we get for " << i << " and " << j << "\n";
      trace_out("There are this many of them : ",locations[index].nr_slopes);
     for (k=0;k<locations[index].nr_slopes;k++) {
	locations[index].post_slopes[k]->print();
      }
#endif
     global_free(actualvecindex[index]);
   }
 }
  global_free(actualvecindex);
*/


  matrix_store = (matrix *) malloc (nr_locations * sizeof(matrix));
  is_orig_matrix = (bool *) malloc (nr_locations * sizeof(bool));
  for (i=0;i<nr_locations;i++) {
    is_orig_matrix[i] = false;
  }
  nr_matrices_in_store = nr_locations;
  for (i=0;i<nr_locns_1;i++) {
    // for each aut 2 location
    for (j=0;j<nr_locns_2;j++) {
      // set up invariant for this location
      int index = i*nr_locns_2 + j;
      set_up_product_invariant(i,j,aut1,aut2);
#ifdef DEBUG_MATRIX_STORE
      trace_out("After set_up_prod ,with ind ", index);
      matrix_store[index].print();
      if (index >0) {trace_out("Matr for ",0); matrix_store[0].print();}
#endif
    }
  }
#ifdef DEBUG_MATRIX_STORE
cout << "The actual 1st nr loc matrices, for invars\n" << flush;
for (i=0;i<nr_locations;i++) {
  matrix_store[i].print();
}
#endif

//  cout << "Doing transitions now \n" << flush;

  // set up transitions, now that most of the location info is there,
  // i.e. slopes, names, and more importantly the invar. matrices
  // already installed in first nr_locations spots in matrix_store
  actualguardindex = (int **) malloc (nr_locations * sizeof (int *));
  actualassindex   = (int **) malloc (nr_locations * sizeof (int *));

  clear_ptr_array(actualguardindex,nr_locations);
  clear_ptr_array(actualassindex,nr_locations);
  // set the info with indices to matrices
  for (i=0;i<nr_locns_1;i++) {
    for (j=0;j<nr_locns_2;j++) {
      // set up stuff.  use indices for matrices ref
      set_up_trans(i,j,aut1,aut2,actualguardindex,actualassindex);
#ifdef DEBUG_TRANS
      { int k;
	trace_out("For index ", i*aut2->nr_locations+j);
	for (k=0;k<locations[i*aut2->nr_locations+j].nr_post_trans;k++) {
	  trace_out("HAHAHA: ", actualguardindex[i*aut2->nr_locations+j][k]);
	  trace_out("HAHAHA: ", actualassindex[i*aut2->nr_locations+j][k]);
//	  { int ijk;
//	    for (ijk=0;ijk<=actualassindex[i*aut2->nr_locations+j][k];ijk++) {
//	      matrix_store[ijk].print();
//	    }
//	  }
	  matrix_store[actualassindex[i*aut2->nr_locations+j][k]].print();
	}
      }
#endif
    }
  }
  // now set the real ptrs, now that the memory has settled.
  for (i=0;i<nr_locns_1;i++) {
    for (j=0;j<nr_locns_2;j++) {
      // now refix the ptrs to matrices, based on the 
      // actualguardindex and actualassindex.
      adjust_ptrs_to_matrices(i,j,aut2,actualguardindex,actualassindex);
    }
  }
  free_actual_indexes(actualguardindex,actualassindex);
/*
  free_slope_info(slope_1_min,aut1->nr_locations);
  free_slope_info(slope_1_max,aut1->nr_locations);
  free_slope_info(slope_2_min,aut2->nr_locations);
  free_slope_info(slope_2_max,aut2->nr_locations);
  free_slope_info(undecl_1,aut1->nr_locations);
  free_slope_info(undecl_2,aut2->nr_locations);
*/

//  cerr << "Number matrices saved / freed in setting up trans = " << nr_freed << "\n" << flush;



  set_location_asap();
} // end set_up_locns





// set up info regarding the locations, except for the transitions
// but we don't want to destroy the current automaton's locations do we?
// also set up vector arrays, and matrix arrays
// only create locations for those product locations which are 
//    control-reachable
void
automaton::set_up_reach_locns(automaton *aut1, automaton *aut2, 
			      int nr_reached, int *is_reached) {
// need to 
// 1) determine how many locations
// 2) allocate space
// 3) set names
// 4) set slopes for each
// 5) set inv matrix
// 6) set backwards indices for easy access to component locations ???
//
  int i,j;
  int **slope_1_min = NULL, **slope_1_max=NULL;
  int **slope_2_min = NULL, **slope_2_max = NULL;
  int ** undecl_1 = NULL, **undecl_2 = NULL;
  int nr_locns_1, nr_locns_2;
  int **actualguardindex, ** actualassindex;  // first index is locn, second
                                              // is transition
  int **actualvecindex;  // index of positions of vectors for each locn
  nr_locns_1 = aut1->nr_locations;
  nr_locns_2 = aut2->nr_locations;
  nr_locations = nr_reached;
  locations = (location *) malloc (nr_locations * sizeof(location));

  // set up empty global list of slopes
  nr_slopes = 0;
  slopes = (slope_vector *) NULL;
  actualvecindex = (int **) malloc (nr_locations * sizeof(int *));
  clear_ptr_array(actualvecindex,nr_locations);


  nr_freed = 0;
  int index;

  // set up each individual location
  // for each aut1 location
  for (i=0;i<nr_locns_1;i++) {
    // for each aut 2 location
    for (j=0;j<nr_locns_2;j++) {
      // place most defaults in position, and ...
      index = find_reach_index(i,j,aut1,aut2,is_reached);
      if (index != -1) {
	locations[index] = location();
	locations[index].locn_id = index;
	// set up names
	locations[index].name = product_name(aut1->locations[i].name,
					       aut2->locations[j].name);

	// set up product post slope matrix
	locations[index].post_slopes_matrix = 
	  aut1->locations[i].post_slopes_matrix->copy()->
	    intersection_no_simp(*(aut2->locations[j].post_slopes_matrix->constraints()));

	locations[index].post_slopes_matrix->fix_strict();
	locations[index].post_slopes_matrix->constraints()->nonstrict =  
	  !locations[index].post_slopes_matrix->strict;
      }
    }
  }


  matrix_store = (matrix *) malloc (nr_locations * sizeof(matrix));
  is_orig_matrix = (bool *) malloc (nr_locations * sizeof(bool));
  for (i=0;i<nr_locations;i++) {
    is_orig_matrix[i] = false;
  }
  nr_matrices_in_store = nr_locations;

  for (i=0;i<nr_locns_1;i++) {
    // for each aut 2 location
    for (j=0;j<nr_locns_2;j++) {
      // set up invariant for this location
      index = find_reach_index(i,j,aut1,aut2,is_reached);
      if (index != -1) {
	set_up_reach_product_invariant(i,j,aut1,aut2, index);
#ifdef DEBUG_MATRIX_STORE
	trace_out("After set_up_prod ,with ind ", index);
	matrix_store[index].print();
	if (index >0) {trace_out("Matr for ",0); matrix_store[0].print();}
#endif
      }
    }
  }

#ifdef DEBUG_MATRIX_STORE
cout << "The actual 1st nr loc matrices, for invars\n" << flush;
for (i=0;i<nr_locations;i++) {
  matrix_store[i].print();
}
#endif

//  cout << "Doing transitions now \n" << flush;

  // set up transitions, now that most of the location info is there,
  // i.e. slopes, names, and more importantly the invar. matrices
  // already installed in first nr_locations spots in matrix_store
  actualguardindex = (int **) malloc (nr_locations * sizeof (int *));
  actualassindex   = (int **) malloc (nr_locations * sizeof (int *));

// up to here **********************************

  clear_ptr_array(actualguardindex,nr_locations);
  clear_ptr_array(actualassindex,nr_locations);
  // set the info with indices to matrices

  for (i=0;i<nr_locns_1;i++) {
    for (j=0;j<nr_locns_2;j++) {
      // set up stuff.  use indices for matrices ref
      index = find_reach_index(i,j,aut1,aut2,is_reached);
      if (index != -1) {
	set_up_reach_trans(i,j,aut1,aut2,actualguardindex,actualassindex, 
			   index,is_reached);
#ifdef DEBUG_TRANS
      { int k;
	trace_out("For index ", i*aut2->nr_locations+j);
	for (k=0;k<locations[i*aut2->nr_locations+j].nr_post_trans;k++) {
	  trace_out("HAHAHA: ", actualguardindex[i*aut2->nr_locations+j][k]);
	  trace_out("HAHAHA: ", actualassindex[i*aut2->nr_locations+j][k]);
	  matrix_store[actualassindex[i*aut2->nr_locations+j][k]].print();
	}
      }
#endif
      }
    }
  }
  // now set the real ptrs, now that the memory has settled.

  for (i=0;i<nr_locns_1;i++) {
    for (j=0;j<nr_locns_2;j++) {
      // now refix the ptrs to matrices, based on the 
      // actualguardindex and actualassindex.
      index = find_reach_index(i,j,aut1,aut2,is_reached);
      if (index != -1) {
	adjust_ptrs_to_matrices_reach(i,j,aut2,actualguardindex,actualassindex,
				      index);
      }
    }
  }
  free_actual_indexes(actualguardindex,actualassindex);

  set_location_asap();
} // end set_up_reach_locns



// return the composition of 2 automata
// don't destroy any of the data/ptrs from the old ones.
// also, make product an independent entity, with no dependent ptrs to
// component automata's free space, etc.
automaton
automaton::compose(automaton aut) {
//
// do not perform reachability analysis now.
// optimize later by avoiding some locations/transitions and building
//    an intermediate automaton which will be converted to C++ aut via
//    dynamic resizing of hashed array space for post_trans, matrix_store
//

  automaton a = automaton();
  
  a.name = product_name(name,aut.name);

  a.set_up_comp_inf(this,&aut);

  // set up alphabet info
  a.set_up_alphabet(this,&aut);

  // set up array of locations
  // will be size |A1|. |A2|
  // store in first_index_pos * nr_in_second + second_index_pos
  a.set_up_locns(this,&aut);

  a.set_up_init_info(this,&aut);

  return (a);
}

// do dfs from index ind, setting is_reached values to 1 whenever reaching
// a new index node (array elts have value -1 for not reached)
// reachability is in terms of control information only
void
reach_dfs(int id1, int id2, 
	  automaton *aut1, automaton *aut2, 
	  int *is_reached) {

  int index = id1 * aut2->nr_locations + id2;

  if (is_reached[index] != -1) {
    return;
  }
  is_reached[index] = 1;

  // now search forward
  bool add_single_trans;
  for (int i=0;i<aut1->locations[id1].nr_post_trans;i++) {
    add_single_trans = false;
    if (has_sync_label(aut1->locations[id1].post_trans[i])) {
      char *label = aut1->locations[id1].post_trans[i].sync_label;
      if (label_in_alphabet(label,aut2->sync_label_names,aut2->nr_sync_labels))
	{
	for (int j = 0;j<aut2->locations[id2].nr_post_trans;j++) {
	  if (has_sync_label(aut2->locations[id2].post_trans[j])) {
	    if (strcmp(label,
		       aut2->locations[id2].post_trans[j].sync_label) == 0) {
	      // add a synchronized transition
	      reach_dfs(aut1->locations[id1].post_trans[i].dest->locn_id,
			aut2->locations[id2].post_trans[j].dest->locn_id,
			aut1,aut2,is_reached);
	    }
	  }
	}
      }
      else {
	// label does not appear in other aut's alphabet
	add_single_trans = true;
      }
    }
    else {
      // no sync label.  just add single transition
      add_single_trans = true;
    }
    if (add_single_trans) {
      // add the transition
       reach_dfs(aut1->locations[id1].post_trans[i].dest->locn_id,
		 id2,
		 aut1,aut2,is_reached);   
     }
  }
  for (int j=0;j<aut2->locations[id2].nr_post_trans;j++) {
    // add if either no sync label, or a sync label not in A1's alphabet
    if (!has_sync_label(aut2->locations[id2].post_trans[j]) ||
	(has_sync_label(aut2->locations[id2].post_trans[j]) &&
	 !label_in_alphabet(aut2->locations[id2].post_trans[j].sync_label,
			   aut1->sync_label_names,aut1->nr_sync_labels))) {
      // add single transition
       reach_dfs(id1,
		 aut2->locations[id2].post_trans[j].dest->locn_id,
		 aut1,aut2,is_reached);  
     }
  }
}


// given two automata to compose, aut1, aut2, 
// set nr_reached to the number of control-reachable locns
//  and is_reached_p to the integer array of whether locns are reached
//  indices as loc_ind_1 * nr_locations(aut2) + loc_ind_2
// value is -1 if not reached, and an index number if reached. 0,1,...
void
find_reached_locns(automaton *aut1, automaton *aut2, 
		   int *nr_reached_p, int **is_reached_p) {
  int i;
  // for now say they all are
  *nr_reached_p = aut1->nr_locations * aut2->nr_locations;
  *is_reached_p = (int *) malloc ( *nr_reached_p * sizeof(int));
/*

  // assume all reached for now
  for (i=0;i<*nr_reached_p;i++)
    (*is_reached_p)[i] = i;

*/

  // use simple dfs
  for (i=0;i<*nr_reached_p;i++)
    (*is_reached_p)[i] = -1;
  
  // find index number of init location.
  reach_dfs(aut1->init_location_p->locn_id,aut2->init_location_p->locn_id,
	    aut1,aut2,*is_reached_p);  

  int index = 0;
  for (i=0;i<*nr_reached_p;i++) {
    if ((*is_reached_p)[i] != -1) {
      (*is_reached_p)[i] = index;
      index++;
    }
  }
  *nr_reached_p = index;

}



// return the composition of 2 automata
// don't destroy any of the data/ptrs from the old ones.
// also, make product an independent entity, with no dependent ptrs to
// component automata's free space, etc.
automaton
automaton::reach_compose(automaton aut) {
//
// do not perform reachability analysis now.
// optimize later by avoiding some locations/transitions and building
//    an intermediate automaton which will be converted to C++ aut via
//    dynamic resizing of hashed array space for post_trans, matrix_store
//

  automaton a = automaton();
  a.name = product_name(name,aut.name);
  a.set_up_comp_inf(this,&aut);
  // set up alphabet info
  a.set_up_alphabet(this,&aut);


  int nr_reached;
  int *is_reached;
  // find_reached_locations
  find_reached_locns(this, &aut, &nr_reached, &is_reached);

  // set up array of locations
  a.set_up_reach_locns(this,&aut, nr_reached, is_reached);

  a.set_up_init_info_reach(this,&aut,is_reached);
 
  return (a);
}

