/* header file for exec_spec.c.  for executing specification statements */
/* use automaton structure, and spec statements parsed in to C structures   */

/*
// Three major functions:   
// 1) setting up and managing the table of regions
// 2) controlling the execution of statements
// 3) evaluating expressions
//  
*/

#ifndef __EXEC_SPEC_H
#define __EXEC_SPEC_H

#include "include.h"
#include "output.h"
#include "basic.h"
#include "predicate.h"
#include "spec.h"

#include "poly.H"          
#include "extrapoly.H"    
#include "automaton.H"
#include "region.H"

#ifdef __cplusplus
extern "C" {
#endif


struct region_table {
  int            size;
  region_array * regions[MAX_REGION_NAMES];
};




void set_up_reg_table(struct region_table *reg_table, 
		      struct region_name_table *reg_name_table);


void execute_statements(struct statement *stmt, 
			struct region_table *reg_table, 
			automaton *aut);

void execute_statement(struct statement *stmt, 
			struct region_table *reg_table, 
		       automaton *aut);

region_array * evaluate_expr(struct expression *expr, 
			     struct region_table *reg_table, 
			     automaton *aut);

#ifdef __cplusplus
}
#endif

#endif
