/*
    get memory (and time?) usage information.
    Apparently not fully compatible with Solaris configurations, which 
       may not offer the getrusage nor getpagesize system calls.
    You may need to comment out the text of function print_max_mem() 
       to compile under Solaris.
*/

#include <sys/time.h>

#ifdef __SOLARIS
#include </usr/ucbinclude/sys/rusage.h>
#else
#include <sys/resource.h>
#endif


#include "getresources.h"

float
float_for_int_div_int_sq(k,basenum)
  int k; int basenum; {
  float temp = k;
  float tempbase = basenum;
  return(temp/tempbase/tempbase);
}

float
seconds_for_microseconds(k)
  int k; {
  return(float_for_int_div_int_sq(k,1000));
}

float
megabytes_for_bytes(k)
  int k; {
  return( float_for_int_div_int_sq(k,1024) );
}

void
print_max_mem() {
/* this is the primary task we're interested in */


#ifdef __HP
/*
  printf("=================================================================\n");
  printf("Sorry. No resource usage information currently available for HPs\n");
  printf("=================================================================\n");
*/
#else

  struct rusage rusagebuffer;
  int    rusage_output;
  int    pagesize = getpagesize();
  int    bytes_used;    
  int    maxrss;

  float utime;
  float utimemicroseconds;
  float stime;
  float stimemicroseconds;
  float tot_time;
  float tot_timemicroseconds;

  rusage_output = getrusage(RUSAGE_SELF,&rusagebuffer);
  maxrss = rusagebuffer.ru_maxrss;

  printf("=================================================================\n");


#ifdef __DEC_ALPHA
  bytes_used = maxrss * 1024;
  printf("Max memory used = %5d kilobytes =              %6.2f MB \n", 
	 maxrss,
	 megabytes_for_bytes(bytes_used,1024)
	 );
#else
  bytes_used = maxrss * pagesize;
  printf("Max memory used = %5d pages = %8d bytes = %6.2f MB \n", 
	 maxrss,
	 bytes_used,
	 megabytes_for_bytes(bytes_used,1024)
	 );
#endif  
  utime = rusagebuffer.ru_utime.tv_sec;
  utimemicroseconds = seconds_for_microseconds(rusagebuffer.ru_utime.tv_usec);
  stime = rusagebuffer.ru_stime.tv_sec;
  stimemicroseconds = seconds_for_microseconds(rusagebuffer.ru_stime.tv_usec);
  tot_time = utime+stime;
  tot_timemicroseconds = utimemicroseconds + stimemicroseconds;

 printf("Time spent      = %10.2fu + %9.2fs = %10.2f sec total\n",
	utime+utimemicroseconds,
	stime+stimemicroseconds,
	tot_time + tot_timemicroseconds);
  printf("=================================================================\n");
/*
 printf(" %di %do %ds\n", 
	 rusagebuffer.ru_inblock,
	 rusagebuffer.ru_oublock,
	 rusagebuffer.ru_nswap
        );
*/
#endif
}
