%p 4096

%{
#include "include.h"
#include "location.h"
#include "output.h"
#include "parser.h"
#include "y.tab.h"
#include "spec.h"
#include "basic.h"
%}





%{
/*
enum { GOTO, IN, SYNC,  WAIT,  WHILE, NUM,  IDENT, ASSIGN,
       RELOP, LOC, END, AUTOMATON, INITIALLY };
int yylval;
*/
%}

%start norm comment

/* regular definitions */
delim      [ \t\n]
ws         {delim}+
digit      [0-9]
letter     [a-zA-Z]
ident      {letter}({letter}|{digit}|[_])*
/* regident   {upperletter}({lowerletter}|{upperletter}|{digit}|[_])* */
number     {digit}+

%%
<norm>[ \t\n]+   ;
<norm>asap       { return ASAP; }
<norm>analog     { yylval.v_type =  V_ANALOG; return ANALOG; }
<norm>automaton  { return AUTOMATON; }
<norm>clock      { yylval.v_type  = V_CLOCK; return CLOCK; }
<norm>do         { return DO; }
<norm>end        { return END; }
<norm>goto       { return GOTO; }
<norm>False      { return P_FALSE; }
<norm>in         { return IN; }
<norm>integrator { yylval.v_type  =  V_INTEGRATOR; return INTEGRATOR; }
<norm>stopwatch  { yylval.v_type  =  V_INTEGRATOR; return INTEGRATOR; }
<norm>loc        { return LOC; }
<norm>parameter  { yylval.v_type  =  V_PARAMETER; return PARAMETER; } 
<norm>discrete   { yylval.v_type  =  V_DISCRETE; return DISCRETE; } 
<norm>sync       { return SYNC; }
<norm>synclabs   { return SYNCLABS; }
<norm>True       { return P_TRUE; }
<norm>inf        { return INFINITY; }
<norm>vars       { return VARS; }
<norm>wait       { return WAIT; } 
<norm>when       { return WHEN; }
<norm>while      { return WHILE; }
<norm>initially  { return INITIALLY; }
<norm>direction  { return DIRECTION; }
<norm>pre        { return P_PRE; }
<norm>post       { return P_POST; }
<norm>eliminate_non_parameters { return ELIM_ALL_VARS; }
<norm>eliminate_variables {return ELIM_VARS; }
<norm>eliminate_all_locations {return ELIM_ALL_LOCNS; }
<norm>eliminate_locations {return ELIM_LOCNS;}
<norm>final      { return FINAL; }
<norm>or         { return P_OR; }
<norm>and        { return P_AND; }
<norm>not        { return P_NOT; }
<norm>weakdiff   { return P_WEAKDIFF; }
<norm>var        { return VAR; }
<norm>region     { return REGION; }
<norm>free       { return FREE; }
<norm>print      { return PRINT; }
<norm>prints     { return PRINTSTRING; }
<norm>printsize  { return PRINTSIZE; }
<norm>iterate    { return ITERATE; }
<norm>from       { return FROM; }
<norm>using      { return USING; }
<norm>reach      { return REACH; }
<norm>endreach   { return ENDREACH; }
<norm>forward    { return FORWARD; }
<norm>backward   { return BACKWARD; }
<norm>all        { return ALL; }
<norm>hide       { return HIDE; }
<norm>endhide    { return ENDHIDE; }
<norm>omit       { return OMIT; }
<norm>non_parameters {return NON_PARAMETERS; }
<norm>locations     {return LOCATIONS; }
<norm>if         { return IF;}
<norm>then       { return THEN; }
<norm>else       { return ELSE; }
<norm>endif      { return ENDIF; }
<norm>while      { return WHILE; }
<norm>endwhile   { return ENDWHILE; }
<norm>empty      { return EMPTY; }
<norm>hull       { return HULL; }
<norm>diff       { return P_DIFF; }
<norm>weakle     { return WEAKLE; }
<norm>weakeq     { return WEAKEQ; }
<norm>weakge     { return WEAKGE; }
<norm>trace      { return TRACE; }
<norm>to         { return TO; }
<norm>{number}   { yylval.integer = atoi((char *)yytext); return INT; }
<norm>{ident}    { yylval.string = str_preserve((char *)yytext);
                  if (string_pos((char *)yytext,reg_name_table.names,reg_name_table.size)==-1)
                     {
                       return(NAME);
                      }
                      else {
                          /* 
                             fprintf(stderr,"found a regname\n");
                             ansi_fflush(stderr);  
                          */
                         return(REGNAME); 
                       }
                   }

<norm>\"[^"]*\"  { /* Kill the last double-quote */
                   yytext[yyleng-1]='\0';
                 
                   yylval.string = str_preserve((char *)yytext+1);
                   return(STRING);
                 }
<norm>":="       { return ASSIGNMENT; }
<norm>"<"        { yylval.integer = LT; return RELOP; }
<norm>">"        { yylval.integer = GT; return RELOP; }
<norm>"<="       { yylval.integer = LE; return RELOP; }
<norm>">="       { yylval.integer = GE; return RELOP; }
<norm>"="        { yylval.integer = EQ; return RELOP; }
<norm>"!="       { yylval.integer = NEQ; return RELOP; }
<norm>"=="       { return EQUALS; }
<norm>"{"        { return '{'; }
<norm>"}"        { return '}'; }
<norm>"("        { return '('; }
<norm>")"        { return ')'; }
<norm>"["        { return '['; }
<norm>"]"        { return ']'; }
<norm>","        { return ','; }
<norm>":"        { return ':'; }
<norm>";"        { return ';'; }
<norm>"'"        { return PRIM; }
<norm>"&"        { return '&'; }
<norm>"|"        { return '|'; }
<norm>"~"        { return '~'; }
<norm>"-"        { return '-'; }
<norm>"+"        { return '+'; }
<norm>"/"        { return '/'; }
"--"			{
			   BEGIN comment;
			}

<comment>(.)		;	/* discard everything within a comment */
<comment>[\n]		{
			  BEGIN norm; /* newline ends a comment */
			}


%%



/* lexinit - starts the Lexical Analyzer off in the right start condition */

void lexinit()
{ /* to fool C++ linking */
    init_region_name_table();
    BEGIN norm;
}


int
yywrap()
{
    if ((yyin != stdin)) {
	close (yyin);
	yyin = stdin;
    };
   return 1;
}
