
#include <fstream.h>
#include "opoly.h"


int pkgetch()
{
	char c = '#';
	do
	    cin >> c;
	while (c == '#');
	if (c == 'q')
		exit(1);
	return c;
}

matrix* nplus1_matrix(const matrix &m)
{
	matrix* np1 = new matrix(m.nbrows+2,m.nbcolumns+1);


/*
 *	np1->p[0] is the 0<=epsilon row
 *      np1->p[1] is the epsilon<=1 row
 */

	np1->p[0][0] = np1->p[1][0] = 1;
	for (int j = 1; j < m.nbcolumns-1; j++)
		np1->p[0][j] = np1->p[1][j] = 0;
	np1->p[0][m.nbcolumns-1] = 1;
	np1->p[1][m.nbcolumns-1] = -1;
	np1->p[0][m.nbcolumns] = 0;
	np1->p[1][m.nbcolumns] = 1;

/*
 *	for each remaining row, add epsilon term if strict
 */

	for (int i = 2; i < m.nbrows+2; i++)
	{
		if (m.p[i-2][0] == STRICT_CONSTANT)
		{
			np1->p[i][0] = 1;
			for (int j = 1; j < m.nbcolumns - 1; j++)
				np1->p[i][j] = m.p[i-2][j];
			np1->p[i][m.nbcolumns - 1] = -1;
			np1->p[i][m.nbcolumns] = m.p[i-2][m.nbcolumns-1];
		}
		else
		{
			for (int j = 0; j < m.nbcolumns - 1; j++)
				np1->p[i][j] = m.p[i-2][j];
			np1->p[i][m.nbcolumns - 1] = 0;
			np1->p[i][m.nbcolumns] = m.p[i-2][m.nbcolumns-1];
		}
	}
        np1->nonstrict = m.nonstrict;
	return np1;
}

/*
int opoly::empty()
{
  // modified HWT 9/22, for strict/nonstrict representations

  strict = constraints()->is_strict();
  if (!strict)  {
// cout << "using non strict emptiness check\n";
    return check_empty(nplus1_form->frame());
  }
  else 
  {
// cout << "using strict emptiness check\n";
add_eps_constraints();
	matrix* f = nplus1_form->frame();
	for (int i = 0; i < f->nbrows; i++)
	{
	  if (f->p[i][f->nbcolumns-1] != 0 && f->p[i][f->nbcolumns-2] > 0)
			return 0;
	}
	return 1;
      }
}
*/


// modified HWT 9/22 to account for eps_pos
lib_polyhedron* opoly::closure(int return_nplus1_form)
{
	/*
	 *
	 * closure returns the n-dimensional closure 
	 * (or (n+1) form depending on input parameter)
	 * of an (n+1)-dimensional opoly
	 *
	 */
         int i,j,k;

	matrix* n1cons = nplus1_form->constraints();
	int nrows = n1cons->nbrows;
	int ncols;
	if (return_nplus1_form) {
	  ncols = n1cons->nbcolumns;
	}
	else {
	  ncols = n1cons->nbcolumns-1;
	}

	/*
	 *
	 * first we discount any rows only constraining epsilon
	 *
	 */

	for (i = 0; i < n1cons->nbrows-1; i++)
	{
		--nrows;
		for ( j = 1; j <= n1cons->nbcolumns-2; j++)
			if ((j != eps_pos) && n1cons->p[i][j])
			{
				++nrows;
				break;
			}
	}

	matrix* m = new matrix(nrows, ncols);
	int row = 0;
	for (i = 0; i < n1cons->nbrows-1; i++)
	{
		for (j = 1; j <= n1cons->nbcolumns-2; j++)
			if ((j != eps_pos) && n1cons->p[i][j])
			{
				/*
				 *
				 * copy this row
				 *
				 */
			   if (return_nplus1_form) {
			     for ( k = 0;k<ncols;k++) 
			       m->p[row][k] = n1cons->p[i][k];
			     // erase the eps_slack entry
			     m->p[row][eps_pos] = 0;
			     ++row;
			     goto i_end;
			   }
			   else {
				for (k = 0; k < eps_pos; k++)
					m->p[row][k] = n1cons->p[i][k];
                                for (k=eps_pos;k<ncols;k++) 
 				  m->p[row][k] = n1cons->p[i][k+1];
				++row;
				goto i_end;
			      }
			}
		i_end: ;
	}


	/*
	 *
	 * add dummy row
	 *
	 */

	m->p[row][0] = m->p[row][ncols-1] = 1;
	for ( i = 1; i <= ncols-2; i++)
		m->p[row][i] = 0;
        m->nonstrict=true;
	return new lib_polyhedron(*m,true);
}

int 
// opoly::slow_contains_point(vector& v)
opoly::slow_contains_point(vector& v)
{
	/*
	 * Assumes vector v is of the form
	 * "a1 a2 ... an b", denoting the
	 * n-vector (a1/b,a2/b,...,an/b).
	 * Returns true iff this vector lies
	 * in the n-dimensional lib_polyhedron
	 * denoted by *this.
	 */

	int denom = v[v.size-1];
	int dim = v.size;
	int j;

	matrix cons0(dim+1,dim+2), cons01(dim+2,dim+2);
	for (int i = 0; i < dim-1; i++)
	{
		cons0.p[i][0] = cons01.p[i][0] = 0;
		for ( j = 0; j < dim-1; j++)
			cons0.p[i][j+1] =
			cons01.p[i][j+1] = (i == j) ? denom : 0;
		cons0.p[i][dim] = cons01.p[i][dim] = 0;
		cons0.p[i][dim+1] = cons01.p[i][dim+1] = -v[i];
	}

	/*
	 * cons0.p[dim-1]  is the epsilon  = 0 row
	 * cons01.p[dim-1] is the epsilon >= 0 row
	 * cons01.p[dim]   is the epsilon <= 1 row
	 *
	 */

	cons0.p[dim-1][0]  = 0;
	cons01.p[dim-1][0] =
	cons01.p[dim][0]   = 1;
	for ( j = 0; j < dim-1; j++)
		cons0.p[dim-1][j+1]  =
		cons01.p[dim-1][j+1] =
		cons01.p[dim][j+1]   = 0;
	cons0.p[dim-1][dim] = cons01.p[dim-1][dim] = 1;
	cons01.p[dim][dim] = -1;
	cons0.p[dim-1][dim+1] =
	cons01.p[dim-1][dim+1] = 0;
	cons01.p[dim][dim+1] = 1;

	/*
	 * cons0.p[dim] is the dummy row 100...001
	 * cons01.p[dim+1] is the dummy row 100...001
	 *
	 */

	cons0.p[dim][0] 	=
	cons0.p[dim][dim+1] 	=
	cons01.p[dim+1][0] 	=
	cons01.p[dim+1][dim+1] = 1;

	for ( j = 1; j <= dim; j++)
		cons0.p[dim][j]    =
		cons01.p[dim+1][j] = 0;

	lib_polyhedron p0(cons0,true), p01(cons01,true);

	p0.intersection(*(nplus1_form->constraints()));
	p01.intersection(*(nplus1_form->constraints()));

	int returnval = !(p01 <= p0);
	p0.free();
	p01.free();
	return returnval;

/*
 *
 * This correct code fails due to library error.
 *
	if (v.size != nplus1_form->constraints()->nbcolumns-2)
	{
		cout << "Dimension mismatch in opoly::contains_point\n";
		exit(1);
	}


	matrix frame01(3,v.size+2);
	matrix frame0(2,v.size+2);

	frame01.p[0][0] =
	frame01.p[1][0] =
	frame0.p[0][0]  = 1;

	for ( j = 0; j < v.size-1; j++)
		frame0.p[0][j+1]  =
		frame01.p[0][j+1] =
		frame01.p[1][j+1] = v[j];

	frame0.p[0][v.size]  =
	frame01.p[0][v.size] = 0;
	frame01.p[1][v.size] = v[v.size-1];

	frame0.p[0][v.size+1]  =
	frame01.p[0][v.size+1] =
	frame01.p[1][v.size+1] = v[v.size-1];

	frame0.p[1][0]         =
	frame0.p[1][v.size+1]  =
	frame01.p[2][0]        =
	frame01.p[2][v.size+1] = 1;


	for (j = 0; j < v.size; j++)
		frame0.p[1][j+1]  =
		frame01.p[2][j+1] = 0;


	cout << "frame0  = " << frame0 << endl;
	cout << "frame01 = " << frame01 << endl;

	pkgetch();


	lib_polyhedron p0(frame0,false);
	lib_polyhedron p01(frame01,false);


	cout << "p0 = " << p0 << endl;
	cout << "p01 = " << p0 << endl;


	int returnval = !(p01 <= p0);
	p0.free();
	p01.free();

	pkgetch();

	return returnval;
 *
 */

}




int 
// opoly::contains_point(vector& v)
opoly::contains_point(vector& v)
{
	/*
	 * Assumes vector v is of the form
	 * "a1 a2 ... an b", denoting the
	 * n-vector (a1/b,a2/b,...,an/b).
	 * Returns true iff this vector lies
	 * in the n-dimensional lib_polyhedron
	 * denoted by *this.
	 */
  int i,j;

  matrix *  constr = constraints();

// constr->print();
// cerr << "in contains pt\n";

  for (i=0;i<constr->nbrows-1;i++) {
    int prod = 0;
    for (j=1;j<constr->nbcolumns-1;j++) {
      if (j!=eps_pos) {
        prod += v[j-1] * constr->p[i][j];
      }
    }
    prod += v[v.size-1] * constr->p[i][constr->nbcolumns-1];
    if (constr->p[i][0] == 0) {
       if (constr->p[i][eps_pos] != 0) {
	 cerr << "Strange constraint detected in polyhedron\n";
	 exit (1);
         return slow_contains_point(v);
       }
       else if (prod != 0)
         return 0;
    }
    else {
      // inequality
      if (constr->p[i][eps_pos] ==0) {
        if (prod <0)
	  return 0;
      }
      else {
	if (constr->p[i][eps_pos] < 0) {
	  // eps is negative, meaning a strict inequality, or eps <= k
	  if (prod<0) {
	    return 0;
	  }
	  else {
	    if (prod==0) {
 	      // check whether it is eps <= k
	      for (j=1;j<constr->nbcolumns-1;j++) {
		if ((constr->p[i][j]!=0) && (j!=eps_pos)) {
		   return 0;
		 }
	      }
	      if (constr->p[i][constr->nbcolumns-1] <=0)  {
		if (constr->p[i][constr->nbcolumns-1] <0) {
		  cerr << "Strange polyhedron constraints\n";
		  exit(1);
		}
		return 0;
	      }
	    }
	  }
	} // end negative eps
	else {
	  // +eps constraint, check whether eps >=0
	    bool weird = false;
	  if ((prod != 0) || (constr->p[i][constr->nbcolumns-1] !=0)) {
	    weird = true;
	  }
	  for (j=1;j<constr->nbcolumns-1;j++) {
	    if ((j!=eps_pos) && (constr->p[i][j]!=0))
	      weird = true;
	  }
	  if (weird) {
	    cerr << "Strange polyhedron constraints\n";
	    exit(1);
	  }
	  // else   eps >= 0
	}
      }
    }
  }      
  return 1;
}


int opoly::operator<=(opoly& op)
{

/*
 *	*this <= op iff closure(*this) <= closure(op) as polyhedra
 *	and for each vertex v of *this with epsilon>0, v in op.
 *	See Halbwachs, Raymond, and Proy
 *		"Verification of linear hybrid systems by means
 *		 of convex approximations"
 */

    strict = constraints()->nplus1_is_strict(eps_pos);
    if (!strict) {
      simplify();
    }
    else {
      add_eps_constraints();
    }
    op.strict = op.constraints()->nplus1_is_strict(eps_pos);
    if (!op.strict) {
      op.simplify();
    }
    else {
      op.add_eps_constraints();
    }

       if (!op.strict) {
// cout << "using nonstrict inclusion\n";
	 return ( (*nplus1_form) <= *(op.nplus1_form) );
       }
// cout << "using strict inclusion\n";


	poly* thiscl = closure();
	poly* opcl   = op.closure();

	if (!((*thiscl) <= *(opcl)))
	{
	        // HWT, changed to avoid mem leak
	        thiscl->free();
		delete thiscl;
	        opcl->free();
		delete opcl;
		return 0;
	}
	thiscl->free();
	opcl->free();
	delete thiscl;
	delete opcl;


       if (!strict) {
	 add_eps_constraints();
// cerr << "adding eps constraints for inclusion\n";
       }

//        cerr << "containment check with strict ineq  \n";

	matrix* _frame = nplus1_form->frame();

	int dim = _frame->nbcolumns-2;
	vector vertex(dim);

	for (int i = 0; i < _frame->nbrows-1; i++)
		if (_frame->p[i][0] == 1 && _frame->p[i][_frame->nbcolumns-1] != 0
			&& _frame->p[i][eps_pos] != 0)
		{
		/*
		 *  	This row of the frame denotes a vertex with epsilon>0.
		 *      Format: "1 a1 a2 ... a(n+1) b" means (a1/b,...,a(n+1)/b)
		 */
                        int ind = 0;
			for (int j = 1; j < dim+2; j++) {
			  if (j!=eps_pos) {
			    vertex[ind] = _frame->p[i][j];
			    ind++;
			  }
			}
		if (!op.contains_point(vertex))
			{
				vertex.free();
				simplify();
				return 0;
			}
		 }
	vertex.free();
	simplify();
	return 1;
}

opoly* opoly::add_elt(vector *v)
{

	vector w(v->size+1);
	for (int i = 0; i < v->size-1; i++)
		w[i] = (*v)[i];
	w[v->size-1] = 0;
	w[v->size] = RIGHT_RAY_CONSTANT;
	nplus1_form->add_elt(&w);
	// HWT, 9/22/95, free w to plug mem leak
        w.free();
	return this;
}


ostream& operator<<(ostream& s, vector& v)
{
	s << "(";
	for (int i = 0; i < v.size-1; i++)
		s << v[i] << ",";
	s << v[v.size-1] << ")";
	return s;
}

ostream& operator<<(ostream& s, matrix& m)
{
	s << m.nbrows << " " << m.nbcolumns << endl;
	for (int i = 0; i < m.nbrows; i++)
	{
		for (int j = 0; j < m.nbcolumns; j++)
			s << m.p[i][j] << " ";
		s << endl;
	}
	return s;
}

ostream& operator<<(ostream& s, lib_polyhedron& p)
{
	s << *(p.constraints());
	return s;
}

istream& operator>>(istream& s, matrix& m)
{
	if (m.p)
		delete m.p;
	int nrows,ncols;
	s >> nrows >> ncols;
	matrix* temp = new matrix(nrows,ncols);
	for (int i = 0; i < nrows; i++)
		for (int j = 0; j < ncols; j++)
			s >> temp->p[i][j];
	m = *temp;
	return s;
}

istream& operator>>(istream&s, vector &v)
{
	if (v.p)
		delete v.p;
	s >> v.size;
	v.p = new int[v.size];
	for (int i = 0; i < v.size; i++)
		s >> v[i];
	return s;
}

ostream& operator<<(ostream& o, opoly& op)
{
	o << *(op.nplus1_form);
	return o;
}

istream& operator>>(istream& i, opoly& op)
{
	if (op.nplus1_form)
		op.nplus1_form->free();
	matrix* m = new matrix;
	i >> *m;
	op.nplus1_form = new lib_polyhedron(*nplus1_matrix(*m),true);
	return i;
}



