#ifndef __POLY_H
#define __POLY_H



#include<stdlib.h>
// #include<stream.h>

// g++ complains for not including stdio.h and iostream.h
#include<stdio.h>
#include<iostream.h>
#include<iomanip.h>
#include<math.h>

#define true 1
#define false 0
#define nil 0

//typedef char bool;

extern int overflow_avoid_mode;

class vector {
public:
  int size;
  int *p;


  vector();
  vector(int s);
  vector(const vector& vec);
  void free();
  void print();

// added HWT, deleted 5/5/95
//  void operator=(const vector& vec);
// added via PKPK, 9/19/95
   int& operator[](int i) {return p[i];}

};


void Vector_Gcd(int *vec, int size, int &g);


class matrix {
  int *p_init;
public:
  int nbrows;
  int nbcolumns;
  int **p;
  bool nonstrict;  // added HWT, 9/26/95, only used in intersections
  matrix();
  matrix(const matrix& m);            // copy constructor
  matrix(int nr, int nc);
  matrix(int nr, int nc, int* tab);
  matrix(int nr);
  void free();
  void print(ostream& outf=cout);
  void print(int, ostream& outf=cout);
  void ptprint(int form, ostream& outf=cout);

  void ptprintrow(int r, char **vnames, bool *nontrivp, int form, ostream& outf=cout);
  void pretty_print(char **names, int form, ostream& outf=cout);
  matrix* share();
  matrix* copy();
  void operator=(const matrix& m);

  // added HWT
  // set to matrix with all rows = 0, except for final row 1 0000 1
  void blank();
  void set_last_row();
  bool eps_present(int eps_pos);
  bool is_strict();  // whether has any strict inequalities, i.e. 2 in first col
  bool nplus1_is_strict(int k); // assumes in nplus1 form, whether has strict ineq., param is pos of eps_slack var
  bool nplus1_def_nonstrict(int k) ; // yields whether n+1 form definitely represents a nonstrict matrix, possibly empty n+1 form.

  // return Gcd of entire matrix of frame's pts
  int gcd_of_frame_matrix();  
  int gcd_of_constr_matrix();

  // divide is boolean, whether to divide by scalar or multiply
  void scale_frame_matrix(int scalar, int divide);
  void scale_constr_matrix(int scalar, int divide);
  void print_simplified();  // simplify matrix, then print it
};


extern matrix* Read_Matrix();
extern vector* Read_Vector();

void my_new_handler();

// extern void transform(matrix mat, matrix *result, bool constraints_wanted);
/* translation constraints <--> frame, according to 'constraint_wanted' */
extern void transform(matrix& m, matrix *n, matrix *result, bool constraints_wanted, int known);
/* the same when a part of the result is already known */

class lib_polyhedron{
  int dimension;
  matrix C;
  matrix F;
  bool frame_available;
  bool constraints_available;
public:
  lib_polyhedron(matrix& mat, bool constraint); // the constructor, from any representation
  lib_polyhedron(lib_polyhedron& p);           // copy constructor
  // ~lib_polyhedron();
  matrix *constraints(); // returns (after possible computation) the constraints
  matrix *frame();       // returns (after possible computation) the frame
  bool empty;
  void print();
  void frame_print();
  void ptprint(ostream& outf=cout);
  lib_polyhedron *var_increment(int var_index); // increment a variable, in both representations
  lib_polyhedron *var_decrement(int var_index); // decrement a variable, in both representations
  lib_polyhedron *var_reset(int var_index);     // reset a variable (only in the frame)
  lib_polyhedron *var_assign(int var_index, vector *exp); 
                                           // assign a variable with a linear expression
  lib_polyhedron *add_elt(vector *elt);        // add a generating element
  lib_polyhedron *intersection(matrix c);  // intersects with new constraints
                                       // n.b. kills old poly

  lib_polyhedron *hull(matrix f);         // convex union with other frame
  void free();
  lib_polyhedron *copy();
  lib_polyhedron *minimize(); // minimization of both representations
  lib_polyhedron *add_a_null_variable(); // one more variable, initialized to 0
  lib_polyhedron *delete_variable(int d); // suppress one variable
  friend lib_polyhedron* empty_lib_polyhedron(int nbcolumns);


  void operator=(lib_polyhedron& p);

  /* added by HWT */
  // set frame invalid, and free memory if frame was being used.
  void invalidate_frame();
  void invalidate_constraints();



              // existentially quantify var vnr
              // out of a constraints with nr_extra_vars (for primed ass. vars)
	      // destroys old poly
  lib_polyhedron *existent(int vnr, int nr_extra_vars);     
              // existentially quantify nr_exist_vars vars, found in indices
	      // destroys old poly
  lib_polyhedron *existent(int nr_exist_vars, int *indices);
	      // destroys old poly
  lib_polyhedron *CopyColumn( int i2, int i3);
	      // destroys old poly
  lib_polyhedron *copy_extra_to_old(int nr_keep_cols, int nr_copy_cols, 
	int *index_p);
	      // returns new poly, original kept intact
  lib_polyhedron *vaporize_variable(int v); // remove column for this variable
                                        // no regard for semantic of var.
                                        // v is var_nr, not column number
	      // returns new poly, original kept intact
  lib_polyhedron *vaporize_variables(int nr_v); // remove column for last
                                        // nr_v vars.
                                        // no regard for semantic of var.
  lib_polyhedron *hull(lib_polyhedron *p); // takes hull, leaving p intact
                                   // but object destroyed

                                   // return fresh copy of poly. with
                                   // added constraint from row rownr of matrix
                                   // M.  Use negated constraint iff 
                                   // negated = true
  lib_polyhedron *add_row(matrix *mp, int rownr, bool negated);
  int dim();  // returns dimension
};

bool operator == (lib_polyhedron& p1, lib_polyhedron& p2);     // comparison
bool operator <= (lib_polyhedron& p1, lib_polyhedron& p2);     // inclusion
lib_polyhedron *widenning (lib_polyhedron p1, lib_polyhedron p2); // standard widenning
lib_polyhedron *limited_widenning (lib_polyhedron p1, lib_polyhedron p2, matrix cont);
                                // widenning upto given constraints
lib_polyhedron *bounded_widenning (lib_polyhedron p1, lib_polyhedron p2, matrix cont);
                                // widenning upto given constraints and intervals

bool check_empty(matrix* F);  // put here so publicly available
#endif
