// #define DEBUG_PREPOST

#include "include.h"
#include "location.h"
#include "output.h"

#include "poly.H"
#include "extrapoly.H"
#include "automaton.H"
#include "region.H"
#include "prepost.H"



int jaka = 0;


/* WARNING: must maintain booleans for validity of constraints and frames */
/* ================================================== */


void
region::restrict_to_invariant() {
  if (locnp->invariant == NULL) {
    cerr << "Null invariant for location \n" << flush;
    locnp->print();
  }
  polyp->intersection(*(locnp->invariant));
}




/* ================================================ */
/* "Major" routines */

/*
// 8/9/95: retiring this routine in favor of method using slope matrices

// add states for cts evolution, do so for first region only.  ignore .next
// returns a completely new region (with ptr to new poly).
// old region is left intact.
// assumes that the region's poly all lies within the invariant.
region *
region::cts_evolve(enum directions dirn) {
// return a new region, don't alter original!
   region *temp = copy();
  slope_vector **slopevecs;
   int i;
   int nb_slopes = locnp->nr_slopes;

   // was   if ((dirn == POST) && (locnp->asap)) {
   // but even in PRE dirn, don't want any cts evolution.
   if (locnp->asap) {
     // just free the temp  region, not the data it was pointing to.
     delete(temp);
     temp = fresh_copy();
     // region may be head of a list, but we want only a single item
     temp->next = (region *) NULL;
     return ( temp );
   }
  // else add time passage "event"

   slopevecs = locnp->get_slope_vectors_for(dirn);

   //  polyp->copy() returns ptr to fresh poly. fresh data
   temp->polyp = polyp->copy();
   for (i=0;i<nb_slopes;i++) {
     // account for the lcm stored at end of every vector
     //   temporarily replace with 0.  this makes it a ray again.
     int this_lcm = slopevecs[i]->p[slopevecs[i]->size-1];
     slopevecs[i]->p[slopevecs[i]->size-1] = 0;
     temp->polyp->add_elt(slopevecs[i]);  //  arg of add_elt is type vector *
     slopevecs[i]->p[slopevecs[i]->size-1] = this_lcm;
   }
   
   temp->restrict_to_invariant();
   temp->next = NULL;
   return(temp);
 }

*/

// add states for cts evolution, do so for first region only.  ignore .next
// returns a completely new region (with ptr to new poly).
// old region is left intact.
// assumes that the region's poly all lies within the invariant.
region *
region::cts_evolve(enum directions dirn) {
// return a new region, don't alter original!
   int i,j;
   region *temp;
   matrix *slope_fr;
   matrix *reg_fr;


#ifdef DEBUG_PREPOST
  cout << "IN CTS EVOLVE\n";
  print();
#endif

   // was   if ((dirn == POST) && (locnp->asap)) {
   // but even in PRE dirn, don't want any cts evolution.
   if (locnp->asap) {
     temp = fresh_copy();
     // region may be head of a list, but we want only a single item
     temp->next = (region *) NULL;
     return ( temp );
   }
  // else add time passage "event"

   temp = copy();
   //  polyp->copy() returns ptr to fresh poly. fresh data
   // removed 9/25. wasting mem.   temp->polyp = polyp->copy();
   slope_fr = 
     (dirn == POST) 
       ? locnp->post_slopes_matrix->frame()
       : locnp->pre_slopes_matrix->frame();
   reg_fr = polyp->frame();

#ifdef DEBUG_PREPOST
   cout << "start with this frame\n";
   reg_fr->print();
  cout << "add this frame\n";
  slope_fr->print();
#endif

   int ncols = reg_fr->nbcolumns;
   matrix *new_fr = new matrix(slope_fr->nbrows + reg_fr->nbrows-1, reg_fr->nbcolumns);
   
   // add rows from original frame
   for (i=0;i<reg_fr->nbrows-1;i++) {
     for (j=0;j<ncols;j++) {
       new_fr->p[i][j] = reg_fr->p[i][j];
     }
   }
   // add new rays (add each row, with 0 in final column for a generator)
   for (i=0;i<slope_fr->nbrows-1;i++) {
     for (j=0;j<ncols-1;j++) {
       new_fr->p[reg_fr->nbrows+i-1][j] = slope_fr->p[i][j];
     }
     // make it a generator, even if was a point
     new_fr->p[reg_fr->nbrows+i-1][ncols-1] = 0;
   }
   new_fr->set_last_row();

// cout << "Reg fr\n"; reg_fr->print();
// cout << "Slope fr\n"; slope_fr->print();  
// cout << "New fr \n"; new_fr.print();
// ansi_fflush();

#ifdef DEBUG_PREPOST
  cout << "made new matrix\n";
  new_fr.print();

#endif

   temp->polyp = new polyhedron(*new_fr,false);

#ifdef DEBUG_PREPOST
  cout << "got new poly\n";
   temp->polyp->print();
   temp->polyp->constraints()->print();
#endif
   temp->restrict_to_invariant();
#ifdef DEBUG_PREPOST
  cout << "after restrict to invariant\n";
   temp->polyp->print();
   temp->polyp->constraints()->print();
#endif


   temp->next = NULL;
   return(temp);
 }



region *
region::cts_evolve_pre() {
// return new region, don't change old one
   return(cts_evolve(PRE));
}


region *
region::cts_evolve_post() {
// return new region, don't change old one
   return( cts_evolve(POST) );
}

// nr_ass_v_old = nr of variables that are reassigned and used as old vars too
// much of this can be statically precomputed
bool
reassigned(int varnr, assignment ass) {
  int i;

  for (i=0;i<ass.nr_vars;i++) {
    if (ass.indices[i] == varnr+1) {
      return (true);
    }
  }
  for (i=0;i<ass.nr_new_vars;i++) {
    if (ass.new_indices[i] == varnr+1) {
      return (true);
    }
  }
  return(false);
}
  
/*  do the assignments, pre and post, for lists, and for single assgt */
// beware: if ditching the effort and returning an empty poly, make sure 
// it is of the correct dimension, not the expanded dimension for assignment
region *
region::single_event_succ(int trans_nr, enum directions dirn) {
  region *regp;   // this will be returned region for event succ
  transition tr;
  assignment assgt;
  int j;


  /* compute successor of region for single transition  trans_nr */

  // get the assignment info

  int nb_orig_cols = polyp->constraints()->nbcolumns;  /* should be num_vars */
  tr = (dirn == PRE) 
          ? locnp->pre_trans[trans_nr]
	  : locnp->post_trans[trans_nr];
  assgt = tr.ass;

  // set-up the location, etc for new region regp.
  regp = fresh_copy();
  regp->locnp = (dirn == PRE)    // src/dest stored in abs forw direction
           ? tr.src
	   : tr.dest;



  /* actually don't need any intersection here for pre() */
  if (dirn == POST) {
    regp->polyp->intersection(*(tr.enabling_constr));
  }

#ifdef DEBUG_PREPOST
cout << "done the intersection with post's enabling constraints\n" << flush;
trace_out("Size of result is ",regp->polyp->constraints()->nbcolumns);
regp->polyp->print(); fflush(stdout);
#endif

  if (((dirn == PRE) || (assgt.nr_vars > 0))
       && !regp->is_empty()  ) {

    /* add extra columns for primed variables */
    // avoid loss of memory
    polyhedron  *tempp; // was  = regp->polyp;
    regp->polyp = Resize_opoly(regp->polyp, nb_orig_cols + assgt.nr_vars);
//    tempp->free();
//    global_free(tempp);

#ifdef DEBUG_PREPOST

cout << "done the resize by adding " << assgt.nr_vars << " vars to give " <<
   nb_orig_cols + assgt.nr_vars << " in total\n";
cout << "dim of new matrix is " << regp->polyp->constraints()->nbcolumns << "\n";
cout << "dim of ass. matrix to intersect with is " << assgt.ass_constraints->nbcolumns << "\n";
cout << "Nr a vars  " << assgt.nr_vars << ", dim a mat " << assgt.ass_constraints->nbcolumns << ", resized " << regp->polyp->constraints()->nbcolumns << "\n" << flush;
cout << " Ass: \n";
assgt.ass_constraints->print();
cout << flush;
cout << "object: \n";
regp->polyp->constraints()->print();
#endif

     if (regp->polyp->empty() ) {
	cerr << "Error in library: arithmetic overflow\n" << flush;
	cout << "Error in library: arithmetic overflow\n" << flush;
	exit(1);
     }
    regp->polyp->intersection(*(assgt.ass_constraints));
#ifdef DEBUG_PREPOST
   cout << "after the intersection w . ass. constraints\n" << flush;
   regp->polyp->nplus1_form->print();

#endif
    if (!regp->is_empty()) {
      // do the existential stuff
      if (dirn == POST) {
	for (j=0;j<assgt.nr_vars;j++) {
#ifdef DEBUG_PREPOST
	  cout << "about to exist for " << j << "'th var, being " << 
	    assgt.indices[j] << "\n" << flush;
	regp->polyp->nplus1_form->print();

#endif
	  regp->polyp->existent(assgt.indices[j],assgt.nr_vars);
#ifdef DEBUG_PREPOST
	cout << "After the exist\n" << flush;
  cout << "regp : " << regp << "\n";
  cout << "polyp : " <<regp->polyp << "\n" << flush;
  cout << "nplus1 form : " << regp->polyp->nplus1_form << "\n" << flush;
  cout << "nplus1 form's frame : " << regp->polyp->nplus1_form->frame() << "\n" << flush;
cout << "constr : " << regp->polyp->constraints() << "\n" << flush;
	regp->polyp->nplus1_form->print();
#endif


	}
      }
      else {
	// dirn == PRE
	// chenged 10/16/95
	// ignore last var, since an NPLUS1_FORM has eps_slack as last var
	for (j=0;j<var_count-1;j++) {
	  // existent. every col which is affected by reassignment
	  if (reassigned(j,assgt)) {
	    // indices  start at 1 for existent's first parameter
	    regp->polyp->existent(j+1,assgt.nr_vars);  
          }
 	}
      }
    }
//    regp->polyp->minimize();  
// #ifdef DEBUG_PREPOST
//     cout << "after minimize\n" << flush;
//     regp->polyp->print();
// #endif
//    regp->polyp->fix_strict();
//    regp->polyp->simplify();

   /* copy new vars back to old positions */
    if (!regp->is_empty()) {
      regp->polyp->copy_extra_to_old(nb_orig_cols,
				     assgt.nr_vars, 
			             assgt.indices);
#ifdef DEBUG_PREPOST
      cout << " --- after copy extra to old ----\n";
      regp->polyp->constraints()->print();
      cout << "--- end copy extra to old ---\n";
#endif
      /* kill the extra cols, index via variables */
      tempp = regp->polyp;
      regp->polyp = regp->polyp->vaporize_variables(assgt.nr_vars);
#ifdef DEBUG_PREPOST
      cout << "after vaporizing vars\n";
      regp->polyp->constraints()->print();
#endif
      tempp->free();
      global_free(tempp);
    }
  }
  if ((dirn == PRE) && !regp->is_empty()) {

#ifdef DEBUG_PREPOST
      cout << " --- before intersecting with enabling ----\n";
      regp->polyp->nplus1_form->print();
      cout << "--- end  ---\n";
  cout << "intersecting with enabling\n";
   tr.enabling_constr->print();
#endif
    regp->polyp->intersection(*(tr.enabling_constr));
#ifdef DEBUG_PREPOST
      cout << " --- after intersecting with enabling ----\n";
      regp->polyp->nplus1_form->print();
      cout << "--- end  ---\n";
#endif
  }
 if (regp->is_empty()) {
   if (regp->polyp->constraints()->nbcolumns != var_count+1) {
     regp->polyp->free();
     global_free(regp->polyp);
     regp->polyp = empty_polyhedron(var_count+2);
   }
 } 
 else {
   regp->restrict_to_invariant();
#ifdef DEBUG_PREPOST
      cout << " --- after intersection with enabling ----\n";
      regp->polyp->print();
      cout << "--- end  ---\n";
#endif
//   regp->polyp->minimize();  
 }
  regp->next = NULL;
  return(regp);
}



region *
region::event_succ(enum directions dirn) {
// return list of successor regions for instantaneous transition
// out of / into a region.
//
  region *temp, *this_ass_succ;
  int nr_trans;
  int i;

  /* initially no successor regions */
  // add dummy record at head of list, since insert must be made into
  // some non-empty list
  temp = NULL;

  nr_trans = (dirn == PRE) 
                ? locnp->nr_pre_trans
		: locnp->nr_post_trans;
  for (i=0;i<nr_trans;i++) {
    /* compute each successor via a transition */
#ifdef DEBUG_PREPOST
cout << "finding single event succ for trans " << i << "\n" << flush;
cout << "************************************\n";
print();
polyp->constraints()->print();
#endif
    this_ass_succ = single_event_succ(i,dirn);
    if (!this_ass_succ->polyp->empty()) {
      // now we insert into the list of successors
      // assign the trans_in and pred_reg_id info
      this_ass_succ->trans_in = (dirn==PRE) 
	                         ? &(locnp->pre_trans[i])
				 : &(locnp->post_trans[i]);
      this_ass_succ->pred_reg_id = reg_id;
      // insert
      if (temp == NULL) {
	temp = this_ass_succ;
      } 
      else {
	temp = temp->insert(this_ass_succ);
      }
    }
  }
#ifdef DEBUG_PREPOST
if (temp != NULL) {
cout << "Found this list of event succ\n" << flush;
temp->print();
cout << "Over\n" << flush;
}
#endif
  return(temp);
}


region *
region::old_pre() {
/*
 return the list of region for pre of reg
  1) add cts evolution.
  2) then add all regions from trans, inserting into head of a list.
*/
  region * evolved_region_p;
   region * temp;
  evolved_region_p = cts_evolve_pre(); // gives new copy
  temp = evolved_region_p->event_succ(PRE);
  // OK to insert since evolved_region_p is a ptr to singleton region
  if (temp != NULL) {
    temp = temp->insert(evolved_region_p);
    evolved_region_p = temp;
  }
  return(evolved_region_p);
}

// don't destroy region
// return completely new list of pre regions.
// HWT 5/23: does not add cts evolution to current region!
region *
region::pre() {
/*
 return the list of region for pre of reg
  1) for each transition, take event succ, and add cts evolution.
*/

  region * temp, *succ, *next_succ;

  temp = (region *) NULL;
  succ = event_succ(PRE);


#ifdef DEBUG_PREPOST
 cout << "top of post, succ b4 cts evolve \n" << flush; 
  if (succ!=NULL) succ->print();
#endif
  while (succ != NULL) {
    // insert the cts evolved region in temp list
    // succ->cts_evolve is a completely fresh poly. presumably non-empty
    //
    if (temp != NULL) {
      temp = temp->insert(succ->cts_evolve(PRE));
    }
    else {
      temp = succ->cts_evolve(PRE);
    }
    next_succ = succ->next;
    succ->free();
//    global_free(succ);
    succ = next_succ;
  }
  return(temp);
}



region *
region::old_post() {
/*
 return the list of region for pre of reg
  1) add cts evolution.
  2) then add all regions from trans, inserting into head of a list.
*/
  region * evolved_region_p;
  region * temp;
  evolved_region_p = cts_evolve_post(); // gives new copy
cout << "\t:::found cts post: \n"; evolved_region_p->print();
cout << "---------------\n" << flush;
  temp = evolved_region_p->event_succ(POST);
  // OK to insert since evolved_region_p is a ptr to singleton region
  if (temp != NULL) {
    temp = temp->insert(evolved_region_p);
    evolved_region_p = temp;
  }
  return(evolved_region_p);
}


// don't destroy region
// return completely new list of post regions.
// HWT 5/23: does not add cts evolution to current region!
region *
region::post() {
/*
 return the list of region for pre of reg
  1) for each transition, take event succ, and add cts evolution.
*/

  region * temp, *succ, *next_succ;

  temp = (region *) NULL;
  succ = event_succ(POST);
#ifdef DEBUG_PREPOST
 cout << "top of post, succ b4 cts evolve \n" << flush; 
  if (succ!=NULL) succ->print();
#endif
  while (succ != NULL) {
    // insert the cts evolved region in temp list
    // succ->cts_evolve is a completely fresh poly. presumably non-empty
    //
    if (temp != NULL) {
      temp = temp->insert(succ->cts_evolve(POST));
    }
    else {
      temp = succ->cts_evolve(POST);
    }
    next_succ = succ->next;
    succ->free();
//    global_free(succ);
    succ = next_succ;
  }
  return(temp);
}





// don't destroy region
// return completely new list of post regions.
// HWT 5/23: does add cts evolution to current region too!
region *
region::full_post() {
  region *reg = post();
  region *thisctspost = cts_evolve(POST);
  thisctspost->next = reg;
  return(thisctspost);
}



// don't destroy region
// return completely new list of pre regions.
// HWT 5/23: does add cts evolution to current region too!
region *
region::full_pre() {
  region *reg = pre();
  region *thisctspre = cts_evolve(PRE);
  thisctspre->next = reg;
  return(thisctspre);
}



