/* header file for parser.  for specification statements (in C) */
/* parse in to these C structures, use directly in C++ part */

#ifndef __SPEC_H
#define __SPEC_H

#include "include.h"
#include "output.h"
#include "basic.h"
#include "predicate.h"

#ifdef __cplusplus
extern "C" {
#endif

#define MAX_STMT_DEPTH 20         /* max nested level */
#define MAX_STMT_STACK_SIZE 200   /* stack includes depth + current list */
#define MAX_EXPR_DEPTH 40
#define MAX_REGION_NAMES 100      /* max number names in table */
#define MAX_BOOLEXPR_DEPTH 100   

struct region_name_table {
  char * names[MAX_REGION_NAMES];
  int    size;
};




enum stmt_type {FREE_STMT, PRINT_STMT, PRINT_STR_STMT, 
		  PRINT_SIMPLIFIED_STMT, PRINTSIZE_STMT,
		  ASSIGN_STMT, IF_THEN_STMT, IF_THEN_ELSE_STMT, 
		  TRACE_STMT,
		  WHILE_STMT
	      };

/* have no explicit type for enclosing an expression in parentheses */
enum expr_type {REGION_NAME, AND, OR, WEAKDIFF, PRE_EXPR, POST_EXPR, 
		  PREDICATE, 
		  REACHFORWARD, REACHBACKWARD, RECURSION, HIDE_EXPR, 
		  OMIT_EXPR, HULL_EXPR,
		  DIFF, COMPLEMENT
	      };
       

enum boolexpr_type { BE_AND, BE_OR, BE_NOT, BE_INEQ, BE_EMPTY,
		   BE_WEAKLE, BE_WEAKEQ, BE_WEAKGE};

struct elim_info {
  int elim_all;  /* = 0 for don't elim all, or non_params, use name_list */
                 /* = 1 for non params, */
                 /* = 2 for all */
  struct var_name *name_list;
};

struct expression {
   enum expr_type      type;
   char *              rname;   
   int                 rindex;  /* index into table of region arrays */
   struct expression * expr1;
   struct expression * expr2;
   struct statement *   stmt;
    struct elim_info  * elim;
  struct convex_predicate *pred; /* a list of convex pred, i.e. disjunction */
 };

struct boolexpression {
  enum boolexpr_type    type;  /* 1 = either true or false, 0 = relative exp */
  struct expression *  left_expr; /* left operand, also doubles as empty exp */
  struct expression *   right_expr; /* right operand */
  struct boolexpression * boolexpr1; /* left boolean op, doubles as NOT exp */
  struct boolexpression *boolexpr2; /* right bool op */
  int                 optype;     /* operator */
};

struct statement {
   enum stmt_type      type;
   char *              rname;  /* also doubles as output string if reqd */
   int                 rindex; /* index into table of region arrays */
   struct expression * expr;
   struct elim_info  * elim;
   struct boolexpression * boolexpr;
   struct statement * loopstmt;
   struct statement *  thenstmt; 
   struct statement *  elsestmt;
   struct statement *  next;
};   

extern struct region_name_table reg_name_table; /* global region table for parser */
extern struct statement *global_stmt; /* the list of all statements */
extern struct statement * stmt_stack          [MAX_STMT_STACK_SIZE];
extern struct expression * expr_stack[MAX_EXPR_DEPTH];
extern struct boolexpression * boolexpr_stack[MAX_BOOLEXPR_DEPTH];
extern int stmt_stack_nr;
extern int expr_depth;
extern int old_spec_style;
extern int boolexpr_depth;


void init_region_name_table();
void store_region_name(char *rname);

struct expression * get_new_expression();
void print_expression(struct expression *expr);
void print_statements(struct statement *stmt);


void init_statements();
void store_statements();
void init_statement(enum stmt_type type);
void store_name_in_stmt(char *name);
void store_outputstr_in_stmt(char *outputstr);
void store_expr_in_stmt();
void init_expression(enum expr_type type);
void store_expression();
void store_name_expression(char *name);
void store_first_expression();
void store_second_expression();
void store_and_expression();
void store_or_expression();
void store_convex_pred_expression();
void store_weakdiff_expression(); 
void store_diff_expression(); 
void store_complement_expression();

void init_elim_loc_info();
void store_elim_loc_info();

void init_elim_var_info();
void store_elim_var_info();

void print_boolexpression(struct boolexpression *boolexpr);
void init_boolexpression(enum boolexpr_type bet); 
void store_ineq_boolexpression();
void store_and_boolexpression();
void store_or_boolexpression();
void store_not_boolexpression();
void store_expr_in_empty_boolexpression();
void store_ineq_type_in_boolexpression(int relop_type);

void store_stmt_in_stmt(enum stmt_type typ);



#ifdef __cplusplus
}
#endif

#endif


